from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Tuple,
    TypeVar,
    Union,
)

from edsnlp.utils.span_getters import SeqStr, SpanFilter

Binding = Tuple[str, Any]


def _check_path(path: str):
    assert [letter.isalnum() or letter == "_" or letter == "." for letter in path], (
        "The label must be a path of valid python identifier to be used as a getter"
        "in the following template: span.[YOUR_LABEL], such as `label_` or `_.negated"
    )
    if path[0].isalpha() or path[0] == "_":
        return "." + path
    return path


def make_binding_getter(qualifier: Union[str, Binding]):
    """
    Make a qualifier getter

    Parameters
    ----------
    qualifier: Union[str, Binding]
        Either one of the following:
        - a path to a nested attributes of the span, such as "qualifier_" or "_.negated"
        - a tuple of (key, value) equality, such as `("_.date.mode", "PASSED")`

    Returns
    -------
    Callable[[Span], bool]
        The qualifier getter
    """
    if isinstance(qualifier, tuple):
        path, value = qualifier
        path = _check_path(path)
        ctx = {"value": value}
        exec(
            f"def getter(span):\n"
            f"    try:\n"
            f"        return span{path} == value\n"
            f"    except AttributeError:\n"
            f"        return False\n",
            ctx,
            ctx,
        )
        return ctx["getter"]
    else:
        path = _check_path(qualifier)
        ctx = {}
        exec(
            f"def getter(span):\n"
            f"    try:\n"
            f"        return span{path}\n"
            f"    except AttributeError:\n"
            f"        return None\n",
            ctx,
            ctx,
        )
        return ctx["getter"]


def make_binding_setter(binding: Binding):
    """
    Make a qualifier setter

    Parameters
    ----------
    binding: Binding
        A pair of
        - a path to a nested attributes of the span, such as `qualifier_` or `_.negated`
        - a value assignment

    Returns
    -------
    Callable[[Span]]
        The qualifier setter
    """
    if isinstance(binding, tuple):
        path, value = binding
        path = _check_path(path)
        fn_string = f"""def setter(span): span{path} = value"""
        ctx = {"value": value}
        exec(fn_string, ctx, ctx)
    else:
        path = _check_path(binding)
        fn_string = f"""def setter(span, value): span{path} = value"""
        ctx = {}
        exec(fn_string, ctx, ctx)
    return ctx["setter"]


K = TypeVar("K")
V = TypeVar("V")


class keydefaultdict(dict):
    def __init__(self, default_factory: Callable[[K], V]):
        super().__init__()
        self.default_factory = default_factory

    def __missing__(self, key: K) -> V:
        ret = self[key] = self.default_factory(key)
        return ret


BINDING_GETTERS = keydefaultdict(make_binding_getter)
BINDING_SETTERS = keydefaultdict(make_binding_setter)

Qualifiers = Dict[str, SpanFilter]


def validate_qualifiers(value: Union[SeqStr, Dict[str, SpanFilter]]) -> Qualifiers:
    if callable(value):
        return value
    if isinstance(value, str):
        return {value: True}
    if isinstance(value, list):
        return {qlf: True for qlf in value}
    elif isinstance(value, dict):
        new_value = {}
        for k, v in value.items():
            if isinstance(v, bool):
                new_value[k] = v
            elif isinstance(v, str):
                new_value[k] = [v]
            elif isinstance(v, (list, tuple)) and all(isinstance(i, str) for i in v):
                new_value[k] = list(v)
            else:
                raise TypeError(
                    f"Invalid entry {value} ({type(value)}) for Qualifiers, "
                    f"expected bool/string(s), dict of bool/string(s) or callable"
                )
        return new_value
    else:
        raise TypeError(
            f"Invalid entry {value} ({type(value)}) for SpanSetterArg, "
            f"expected bool/string(s), dict of bool/string(s) or callable"
        )


class QualifiersArg:
    """
    Valid values for the `qualifiers` argument of a component can be :

    - a (span) -> qualifier callable
    - a qualifier name ("_.negated")
    - a list of qualifier names (["_.negated", "_.event"])
    - a dict of qualifier name to True or list of labels, to filter the qualifiers

    Examples
    --------
    - `qualifiers="_.negated"` will use the `negated` extention of the span
    - `qualifiers=["_.negated", "_.past"]` will use the `negated` and `past`
       extensions of the span
    - `qualifiers={"_.negated": True, "_.past": "DATE"}` will use the `negated`
       extension of any span, and the `past` extension of spans with the `DATE` label
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, config=None) -> Qualifiers:
        return validate_qualifiers(value)


if TYPE_CHECKING:
    QualifiersArg = Union[SeqStr, Dict[str, Union[bool, SeqStr]], Callable]  # noqa: F811



from edsnlp.core import registry
from edsnlp.pipes.core.contextual_matcher import ContextualMatcher

DEFAULT_CONFIG = dict(
    assign_as_span=False,
    alignment_mode="expand",
    attr="NORM",
    regex_flags=0,
    ignore_excluded=False,
    ignore_space_tokens=False,
    include_assigned=False,
    label_name=None,
    label=None,
    span_setter={"ents": True},
)

create_component = registry.factory.register(
    "eds.contextual-matcher",
    deprecated=["contextual-matcher"],
)(ContextualMatcher)



from typing import TYPE_CHECKING, Any, Generic, List, TypeVar

import pydantic
from confit.errors import patch_errors

T = TypeVar("T")


class MetaAsList(type):
    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        cls.item = Any

    def __getitem__(self, item):
        new_type = MetaAsList(self.__name__, (self,), {})
        new_type.item = item
        return new_type

    def validate(cls, value, config=None):
        if isinstance(value, dict):
            value = [value]
        if not isinstance(value, list):
            value = [value]
        try:
            return pydantic.parse_obj_as(List[cls.item], value)
        except pydantic.ValidationError as e:
            e = patch_errors(e, drop_names=("__root__",))
            e.model = cls
            raise e

    def __get_validators__(cls):
        yield cls.validate


class AsList(Generic[T], metaclass=MetaAsList):
    pass


if TYPE_CHECKING:
    AsList = List[T]  # noqa: F811



"""
EDS-NLP
"""

import sys
import importlib.abc
import importlib.util
from pathlib import Path
from spacy import pipeline as _spacy_pipeline  # noqa: F401
from . import extensions
from . import patch_spacy
from .core.pipeline import Pipeline, blank, load
from .core.registries import registry
import edsnlp.data  # noqa: F401
import edsnlp.pipes

__version__ = "0.11.1"

BASE_DIR = Path(__file__).parent


# Everything below is to support deprecated use of edsnlp.pipelines
# route imports of submodules of edsnlp.pipelines to their edsnlp.pipes counterparts

class AliasPathFinder(importlib.abc.MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        if fullname.startswith("edsnlp.pipelines"):
            new_name = "edsnlp.pipes" + fullname[16:]
            spec = importlib.util.spec_from_loader(fullname, AliasLoader(new_name))
            return spec


class AliasLoader(importlib.abc.Loader):
    def __init__(self, alias):
        self.alias = alias

    def create_module(self, spec):
        # import the edsnlp.pipe.* module first, then alias it to edsnlp.pipelines.*
        module = importlib.import_module(self.alias)
        sys.modules[spec.name] = module
        return module

    def exec_module(self, module):
        pass


# Add the custom finder to sys.path_hooks
sys.meta_path.insert(0, AliasPathFinder())



import re
from typing import Any, Callable, List, Optional, Union

from edsnlp.core import PipelineProtocol, registry
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.ner.scores.base_score import SimpleScoreMatcher
from edsnlp.pipes.ner.scores.charlson import patterns

DEFAULT_CONFIG = dict(
    regex=patterns.regex,
    value_extract=patterns.value_extract,
    score_normalization=patterns.score_normalization_str,
    attr="NORM",
    window=7,
    ignore_excluded=False,
    ignore_space_tokens=False,
    flags=0,
    label="charlson",
    span_setter={"ents": True, "charlson": True},
)


@registry.factory.register(
    "eds.charlson",
    assigns=["doc.ents", "doc.spans"],
    deprecated=[
        "eds.charlson",
        "charlson",
    ],
)
def create_component(
    nlp: PipelineProtocol,
    name: Optional[str] = "charlson",
    *,
    regex: List[str] = patterns.regex,
    value_extract: str = patterns.value_extract,
    score_normalization: Union[
        str, Callable[[Union[str, None]], Any]
    ] = patterns.score_normalization_str,
    attr: str = "NORM",
    window: int = 7,
    ignore_excluded: bool = False,
    ignore_space_tokens: bool = False,
    flags: Union[re.RegexFlag, int] = 0,
    label: str = "charlson",
    span_setter: SpanSetterArg = {"ents": True, "charlson": True},
):
    '''
    The `eds.charlson` component extracts the
    [Charlson Comorbidity Index](https://www.mdcalc.com/charlson-comorbidity-index-cci).

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.charlson())

    text = """
    Charlson à l'admission: 7.
    Charlson:
    OMS:
    """

    doc = nlp(text)
    doc.ents
    # Out: (Charlson à l'admission: 7,)
    ```

    We can see that only one occurrence was extracted. The second mention of
    Charlson in the text doesn't contain any numerical value, so it isn't extracted.

    Extensions
    ----------
    Each extraction exposes 2 extensions:

    ```python
    ent = doc.ents[0]

    ent._.score_name
    # Out: 'charlson'

    ent._.score_value
    # Out: 7
    ```

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : Optional[str]
        Name of the component
    regex : List[str]
        A list of regexes to identify the score
    attr : str
        Whether to match on the text ('TEXT') or on the normalized text ('NORM')
    value_extract : str
        Regex with capturing group to get the score value
    score_normalization : Union[str, Callable[[Union[str,None]], Any]]
        Function that takes the "raw" value extracted from the `value_extract`
        regex and should return:

        - None if no score could be extracted
        - The desired score value else
    window : int
        Number of token to include after the score's mention to find the
        score's value
    ignore_excluded : bool
        Whether to ignore excluded spans when matching
    ignore_space_tokens : bool
        Whether to ignore space tokens when matching
    flags : Union[re.RegexFlag, int]
        Regex flags to use when matching
    label : str
        Label name to use for the `Span` object and the extension
    span_setter: SpanSetterArg
        How to set matches on the doc

    Returns
    -------
    SimpleScoreMatcher
    '''
    return SimpleScoreMatcher(
        nlp,
        name=name,
        regex=regex,
        value_extract=value_extract,
        score_normalization=score_normalization,
        attr=attr,
        window=window,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
        flags=flags,
        label=label,
        span_setter=span_setter,
    )



from edsnlp.core import registry

from .family import FamilyContextQualifier

DEFAULT_CONFIG = dict(
    attr="NORM",
    family=None,
    termination=None,
    use_sections=True,
    span_getter=None,
    on_ents_only=True,
    explain=False,
)

create_component = registry.factory.register(
    "eds.family",
    assigns=["span._.family"],
    deprecated=["family"],
)(FamilyContextQualifier)



import pytest
from pytest import fixture

from edsnlp.pipelines.core.matcher import GenericMatcher


@fixture
def nlp(blank_nlp):
    blank_nlp.add_pipe("eds.normalizer")
    return blank_nlp


@fixture
def doc(nlp, text):
    return nlp(text)


@fixture
def matcher_factory(nlp):
    default_config = dict(
        attr="TEXT",
        ignore_excluded=False,
    )

    def factory(terms=None, regex=None, **kwargs):
        assert terms or regex

        config = dict(**default_config)
        config.update(kwargs)

        return GenericMatcher(
            nlp=nlp,
            terms=terms,
            regex=regex,
            **config,
        )

    return factory


def test_matcher_config_typo(nlp):
    with pytest.raises(ValueError):
        nlp.add_pipe(
            "matcher",
            config={
                "terms": {"test": ["test"]},
                "term_matcher": "exoct",
            },
        )


def test_exact_matcher_spacy_factory(nlp):
    nlp.add_pipe(
        "matcher",
        config={
            "terms": {"test": ["test"]},
            "term_matcher": "exact",
        },
    )


def test_simstring_matcher_spacy_factory(nlp):
    nlp.add_pipe(
        "matcher",
        config={
            "terms": {"test": ["test"]},
            "term_matcher": "simstring",
            "term_matcher_config": {
                "measure": "dice",
            },
        },
    )


def test_terms(doc, matcher_factory):
    matcher = matcher_factory(
        terms=dict(patient="patient", anomalie="anomalie"),
        attr="NORM",
    )
    doc = matcher(doc)
    assert len(doc.ents) == 3, "There should be two entities."


def test_regex(doc, matcher_factory):
    matcher = matcher_factory(
        regex=dict(patient=r"patient", anomalie=r"anomalie"),
        attr="TEXT",
    )
    doc = matcher(doc)
    assert len(doc.ents) == 3, "There should be two entities."


def test_space(doc, matcher_factory):
    matcher = matcher_factory(
        terms=dict(holidays=r"vacances d'été"),
        attr="NORM",
        ignore_space_tokens=True,
    )
    doc = matcher(doc)
    assert len(doc.ents) == 1, "There should be one entity."



import re
from typing import Union

import spacy

regex = [r"[Ee]lston (& |et |and )?[Ee]llis", r"\b[Ee]{2}\b"]

pattern1 = r"[^\d\(\)]*[0-3]"
pattern2 = r".{0,2}[\+,]"
value_extract = rf"(?s).(\({pattern1}{pattern2}{pattern1}{pattern2}{pattern1}\))"

score_normalization_str = "score_normalization.elstonellis"


@spacy.registry.misc(score_normalization_str)
def score_normalization(extracted_score: Union[str, None]):
    """
    Elston and Ellis score normalization.
    If available, returns the integer value of the Elston and Ellis score.
    """
    try:
        x = 0
        for i in re.findall(r"[0-3]", extracted_score):
            x += int(i)

        if x <= 5:
            return 1

        elif x <= 7:
            return 2

        else:
            return 3

    except ValueError:
        return None



import contextlib
import pprint
import warnings
from array import array

from pyspark.sql.types import (
    ArrayType,
    DataType,
    DecimalType,
    NullType,
    StructField,
    StructType,
    _array_type_mappings,
    _infer_schema,
    _infer_type,
    _type_mappings,
)


class PySparkPrettyPrinter(pprint.PrettyPrinter):
    def __init__(self, *args, width=1, **kwargs):
        width = 1
        super().__init__(*args, width=width, **kwargs)

    def _format(self, object, stream, indent, allowance, context, level):
        if isinstance(object, DataType):
            self._pprint_pyspark_type(object, stream, indent, allowance, context, level)
        else:
            super()._format(object, stream, indent, allowance, context, level)

    _dispatch = pprint.PrettyPrinter._dispatch.copy()

    def _pprint_pyspark_type(self, object, stream, indent, allowance, context, level):
        if isinstance(object, StructType):
            stream.write("T.StructType([\n")
            for i, field in enumerate(object.fields):
                if i > 0:
                    stream.write(",\n")
                stream.write(" " * (indent + 4))  # Increase indent for nested elements
                self._pprint_pyspark_type(
                    field, stream, indent + 4, allowance, context, level + 1
                )
            stream.write("\n" + " " * indent + "])")
        elif isinstance(object, StructField):
            stream.write(f"T.StructField('{object.name}', ")
            self._pprint_pyspark_type(
                object.dataType, stream, indent, allowance, context, level + 1
            )
            stream.write(f", {object.nullable})")
        elif isinstance(object, ArrayType):
            stream.write("T.ArrayType(")
            self._pprint_pyspark_type(
                object.elementType, stream, indent, allowance, context, level + 1
            )
            stream.write(f", {object.containsNull})")
        else:
            stream.write(f"T.{type(object).__name__}()")

    for dtype in _type_mappings.values():
        _dispatch[dtype.__repr__] = _pprint_pyspark_type
    _dispatch[DataType.__repr__] = _pprint_pyspark_type


def schema_warning(schema):
    warnings.warn(
        "The following schema was inferred from the dataframe. You should pass "
        "it to `edsnlp.data.to_spark` (and maybe fix it) to speed up the process "
        "and avoid surprises:\n"
        "import pyspark.sql.types as T\n"
        "dtypes = " + PySparkPrettyPrinter().pformat(schema),
        Warning,
    )


def infer_type(obj, **kwargs):
    """Infer the DataType from obj"""
    if obj is None:
        return NullType()

    if hasattr(obj, "__UDT__"):
        return obj.__UDT__

    dataType = _type_mappings.get(type(obj))
    if dataType is DecimalType:
        # the precision and scale of `obj` may be different from row to row.
        return DecimalType(38, 18)
    elif dataType is not None:
        return dataType()

    # if isinstance(obj, dict):
    #     for key, value in obj.items():
    #         if key is not None and value is not None:
    #             return MapType(_infer_type(key), _infer_type(value), True)
    #     return MapType(NullType(), NullType(), True)
    # elif isinstance(obj, list):
    if isinstance(obj, list):
        for v in obj:
            if v is not None:
                return ArrayType(_infer_type(obj[0]), True)
        return ArrayType(NullType(), True)
    elif isinstance(obj, array):
        if obj.typecode in _array_type_mappings:
            return ArrayType(_array_type_mappings[obj.typecode](), False)
        else:
            raise TypeError("not supported type: array(%s)" % obj.typecode)
    else:
        try:
            return _infer_schema(obj)
        except TypeError:
            raise TypeError("not supported type: %s" % type(obj))


def infer_schema(row, names=None, **kwargs):
    """Infer the schema from dict/namedtuple/object"""
    if isinstance(row, dict):
        # PATCHED: don't sort the items.
        # items = sorted(row.items())
        items = list(row.items())

    elif isinstance(row, (tuple, list)):
        if hasattr(row, "__fields__"):  # Row
            items = zip(row.__fields__, tuple(row))
        elif hasattr(row, "_fields"):  # namedtuple
            items = zip(row._fields, tuple(row))
        else:
            if names is None:
                names = ["_%d" % i for i in range(1, len(row) + 1)]
            elif len(names) < len(row):
                names.extend("_%d" % i for i in range(len(names) + 1, len(row) + 1))
            items = zip(names, row)

    elif hasattr(row, "__dict__"):  # object
        items = list(row.__dict__.items())

    else:
        raise TypeError("Can not infer schema for type: %s" % type(row))

    fields = []
    for k, v in items:
        try:
            fields.append(StructField(k, _infer_type(v), True))
        except TypeError as e:
            raise TypeError(
                "Unable to infer the type of the field {}.".format(k)
            ) from e
    return StructType(fields)


@contextlib.contextmanager
def spark_interpret_dicts_as_rows():
    # Replace the code object inside _infer_type
    # with a code object that has the same bytecode
    old_infer_type_code = _infer_type.__code__
    old_infer_schema_code = _infer_type.__code__
    old_infer_type_defaults = _infer_type.__defaults__
    old_infer_schema_defaults = _infer_schema.__defaults__

    _infer_type.__code__ = infer_type.__code__
    _infer_schema.__code__ = infer_schema.__code__
    _infer_type.__defaults__ = infer_type.__defaults__
    _infer_schema.__defaults__ = infer_schema.__defaults__

    yield

    _infer_type.__code__ = old_infer_type_code
    _infer_schema.__code__ = old_infer_schema_code
    _infer_type.__defaults__ = old_infer_type_defaults
    _infer_schema.__defaults__ = old_infer_schema_defaults



from edsnlp.core import registry

from .patterns import default_patterns
from .solid_tumor import SolidTumorMatcher

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    use_tnm=False,
    label="solid_tumor",
    span_setter={"ents": True, "solid_tumor": True},
)

create_component = registry.factory.register(
    "eds.solid_tumor",
    assigns=["doc.ents", "doc.spans"],
)(SolidTumorMatcher)



import inspect
import types
from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Dict, Iterable, Optional, Sequence
from weakref import WeakKeyDictionary

import catalogue
import spacy
from confit import Config, Registry, RegistryCollection, set_default_registry
from confit.errors import ConfitValidationError, patch_errors
from spacy.pipe_analysis import validate_attrs

import edsnlp
from edsnlp.utils.collections import FrozenDict, FrozenList

PIPE_META = WeakKeyDictionary()


def accepted_arguments(
    func: Callable,
    args: Sequence[str],
) -> Sequence[str]:
    """
    Checks that a function accepts a list of keyword arguments

    Parameters
    ----------
    func: Callable[..., T]
        Function to check
    args: Union[str, Sequence[str]]
        Argument or list of arguments to check

    Returns
    -------
    List[str]
    """
    sig = inspect.signature(func)
    has_kwargs = any(
        param.kind == param.VAR_KEYWORD for param in sig.parameters.values()
    )
    if has_kwargs:
        return args
    return [arg for arg in args if arg in sig.parameters]


@dataclass
class FactoryMeta:
    assigns: Iterable[str]
    requires: Iterable[str]
    retokenizes: bool
    default_config: Dict


class CurriedFactory:
    def __init__(self, func, kwargs):
        self.kwargs = kwargs
        self.factory = func
        self.instantiated = None
        self.error = None

    def instantiate(
        obj: Any,
        nlp: "edsnlp.Pipeline",
        path: Optional[Sequence[str]] = (),
    ):
        """
        To ensure compatibility with spaCy's API, we need to support
        passing in the nlp object and name to factories. Since they can be
        nested, we need to add them to every factory in the config.
        """
        if isinstance(obj, CurriedFactory):
            if obj.error is not None:
                raise obj.error

            if obj.instantiated is not None:
                return obj.instantiated

            name = path[0] if len(path) == 1 else None
            parameters = (
                inspect.signature(obj.factory.__init__).parameters
                if isinstance(obj.factory, type)
                else inspect.signature(obj.factory).parameters
            )
            kwargs = {
                key: CurriedFactory.instantiate(
                    obj=value,
                    nlp=nlp,
                    path=(*path, key),
                )
                for key, value in obj.kwargs.items()
            }
            try:
                if nlp and "nlp" in parameters:
                    kwargs["nlp"] = nlp
                if name and "name" in parameters:
                    kwargs["name"] = name
                obj.instantiated = obj.factory(**kwargs)
            except ConfitValidationError as e:
                obj.error = e
                raise ConfitValidationError(
                    patch_errors(e.raw_errors, path, model=e.model),
                    model=e.model,
                    name=obj.factory.__module__ + "." + obj.factory.__qualname__,
                )  # .with_traceback(None)
            # except Exception as e:
            #     obj.error = e
            #     raise ConfitValidationError([ErrorWrapper(e, path)])
            return obj.instantiated
        elif isinstance(obj, dict):
            instantiated = {}
            errors = []
            for key, value in obj.items():
                try:
                    instantiated[key] = CurriedFactory.instantiate(
                        obj=value,
                        nlp=nlp,
                        path=(*path, key),
                    )
                except ConfitValidationError as e:
                    errors.extend(e.raw_errors)
            if errors:
                raise ConfitValidationError(errors)
            return instantiated
        elif isinstance(obj, (tuple, list)):
            instantiated = []
            errors = []
            for i, value in enumerate(obj):
                try:
                    instantiated.append(
                        CurriedFactory.instantiate(value, nlp, (*path, str(i)))
                    )
                except ConfitValidationError as e:  # pragma: no cover
                    errors.append(e.raw_errors)
            if errors:
                raise ConfitValidationError(errors)
            return type(obj)(instantiated)
        else:
            return obj

    def _raise_curried_factory_error(self):
        raise TypeError(
            f"This component ({self}) has not been instantiated yet, likely because it "
            f"was missing an `nlp` pipeline argument. You should either:\n"
            f"- add it to a pipeline: `pipe = nlp.add_pipe(pipe)`\n"
            f"- or fill its `nlp` argument: `pipe = factory(nlp=nlp, ...)`"
        )

    def __call__(self, *args, **kwargs):
        self._raise_curried_factory_error()

    def __getattr__(self, name):
        if name.startswith("__"):
            raise AttributeError(name)
        self._raise_curried_factory_error()


glob = []


def get_all_available_factories(namespaces):
    result = set()
    for ns in namespaces:
        entry_point_ns = "_".join(ns)
        eps = (
            catalogue.AVAILABLE_ENTRY_POINTS.select(group=entry_point_ns)
            if hasattr(catalogue.AVAILABLE_ENTRY_POINTS, "select")
            else catalogue.AVAILABLE_ENTRY_POINTS.get(entry_point_ns, [])
        )
        result.update([ep.name for ep in eps])
        for keys in catalogue.REGISTRY.copy().keys():
            if ns == keys[: len(ns)]:
                result.add(keys[-1])
    return sorted(result)


class FactoryRegistry(Registry):
    """
    A registry that validates the input arguments of the registered functions.
    """

    def get(self, name: str) -> Any:
        """
        Get the registered function for a given name.
        Since we want to be able to get functions registered under spacy namespace,
        and functions defined in memory but not accessible via entry points (spacy
        internal namespace), we need to check multiple registries.

        The strategy is the following:

        1. If the function exists in the edsnlp_factories namespace, return it as a
           curried function.
        2. Otherwise, check spacy namespaces and re-register it under edsnlp_factories
           namespace, then return it as a curried function.
        3. Otherwise, search in edsnlp's entry points and redo steps 1 & 2
        4. Otherwise, search in spacy's points and redo steps 1 & 2
        5. Fail

        name (str): The name.
        RETURNS (Any): The registered function.
        """

        registry_path = list(self.namespace) + [name]
        registry_spacy_path = ["spacy", "factories", name]
        registry_internal_spacy_path = ["spacy", "internal_factories", name]

        def check_and_return():
            if not catalogue.check_exists(*registry_path):
                for path in (registry_spacy_path, registry_internal_spacy_path):
                    if catalogue.check_exists(*path):
                        func = catalogue._get(path)
                        meta = spacy.Language.get_factory_meta(name)

                        self.register(
                            name,
                            func=func,
                            assigns=meta.assigns,
                            requires=meta.requires,
                            retokenizes=meta.retokenizes,
                            default_config=meta.default_config,
                        )

            if catalogue.check_exists(*registry_path):
                func = catalogue._get(registry_path)
                return lambda **kwargs: CurriedFactory(func, kwargs=kwargs)

        # Steps 1 & 2
        func = check_and_return()
        if func is None and self.entry_points:
            # Otherwise, step 3
            self.get_entry_point(name)
            # Then redo steps 1 & 2
            func = check_and_return()
        if func is None:
            # Otherwise, step 4
            if hasattr(spacy.registry, "_entry_point_factories"):
                spacy.registry._entry_point_factories.get_entry_point(name)
                # Then redo steps 1 & 2
                func = check_and_return()

        if func is not None:
            return func

        raise catalogue.RegistryError(
            (
                "Can't find '{name}' in registry {current_namespace}. "
                "Available names: {available_str}"
            ).format(
                name=name,
                current_namespace=" -> ".join(self.namespace),
                available_str=", ".join(
                    get_all_available_factories(
                        [
                            self.namespace,
                            ("spacy", "factories"),
                            ("spacy", "internal_factories"),
                        ]
                    )
                )
                or "none",
            )
        )

    def register(
        self,
        name: str,
        *,
        func: Optional[catalogue.InFunc] = None,
        default_config: Dict[str, Any] = FrozenDict(),
        assigns: Iterable[str] = FrozenList(),
        requires: Iterable[str] = FrozenList(),
        retokenizes: bool = False,
        default_score_weights: Dict[str, Optional[float]] = FrozenDict(),
        invoker: Callable = None,
        deprecated: Sequence[str] = (),
        spacy_compatible: bool = True,
    ) -> Callable[[catalogue.InFunc], catalogue.InFunc]:
        """
        This is a convenience wrapper around `confit.Registry.register`, that
        curries the function to be registered, allowing to instantiate the class
        later once `nlp` and `name` are known.

        Parameters
        ----------
        name: str
        func: Optional[catalogue.InFunc]
        default_config: Dict[str, Any]
        assigns: Iterable[str]
        requires: Iterable[str]
        retokenizes: bool
        default_score_weights: Dict[str, Optional[float]]
        invoker: Callable
        deprecated: Sequence[str]
        spacy_compatible: bool

        Returns
        -------
        Callable[[catalogue.InFunc], catalogue.InFunc]
        """
        save_params = {"@factory": name}

        def register(fn: catalogue.InFunc) -> catalogue.InFunc:
            assert (
                not spacy_compatible
                or len(accepted_arguments(fn, ["nlp", "name"])) == 2
            ), (
                "Spacy compatible factories functions must accept nlp and name as "
                "arguments. Either set register(..., spacy_compatible=False) or "
                "add nlp and name to the arguments."
            )

            meta = FactoryMeta(
                assigns=validate_attrs(assigns),
                requires=validate_attrs(requires),
                retokenizes=retokenizes,
                default_config=default_config,
            )

            def invoke(validated_fn, kwargs):
                if default_config is not None:
                    kwargs = (
                        Config(default_config)
                        .resolve(registry=self.registry)
                        .merge(kwargs)
                    )
                instantiated = (
                    invoker(validated_fn, kwargs)
                    if invoker is not None
                    else validated_fn(kwargs)
                )
                PIPE_META[instantiated] = meta
                return instantiated

            annotations = (
                fn.__annotations__
                if not isinstance(fn, type)
                else fn.__init__.__annotations__
            )
            old_annotation = None
            if "nlp" in annotations:  # annotations["nlp"] is spacy.Language:
                old_annotation = annotations["nlp"]
                annotations["nlp"] = Optional[edsnlp.core.PipelineProtocol]
            registered_fn = Registry.register(
                self,
                name=name,
                save_params=save_params,
                skip_save_params=["nlp", "name"],
                func=fn,
                invoker=invoke,
                deprecated=deprecated,
            )
            if old_annotation is not None:
                annotations["nlp"] = old_annotation

            if spacy_compatible:
                for spacy_pipe_name in (name, *deprecated):
                    spacy.Language.factory(
                        name=spacy_pipe_name,
                        default_config=default_config,
                        assigns=assigns,
                        requires=requires,
                        default_score_weights=default_score_weights,
                        retokenizes=retokenizes,
                        func=registered_fn,
                    )

            @wraps(fn)
            def curried_registered_fn(
                nlp=inspect.Signature.empty,
                name=inspect.Signature.empty,
                **kwargs,
            ):
                sig = inspect.signature(fn)
                if nlp is not inspect.Signature.empty:
                    kwargs["nlp"] = nlp
                if name is not inspect.Signature.empty:
                    kwargs["name"] = name
                bound = sig.bind_partial(**kwargs)
                if (
                    "nlp" in sig.parameters
                    and sig.parameters["nlp"].default is sig.empty
                    and bound.arguments.get("nlp", sig.empty) is sig.empty
                ):
                    return CurriedFactory(registered_fn, dict(bound.arguments))
                return registered_fn(**bound.arguments)

            return (
                curried_registered_fn
                if isinstance(fn, types.FunctionType)
                else registered_fn
            )

        return register(func) if func is not None else register


class registry(RegistryCollection):
    factories = FactoryRegistry(("edsnlp", "factories"), entry_points=True)
    factory = factories
    misc = Registry(("spacy", "misc"), entry_points=True)
    languages = Registry(("spacy", "languages"), entry_points=True)
    tokenizers = Registry(("spacy", "tokenizers"), entry_points=True)
    scorers = Registry(("spacy", "scorers"), entry_points=True)
    accelerator = Registry(("edsnlp", "accelerator"), entry_points=True)
    adapters = Registry(("edsnlp", "adapters"), entry_points=True)
    readers = Registry(("edsnlp", "readers"), entry_points=True)
    writers = Registry(("edsnlp", "writers"), entry_points=True)


set_default_registry(registry)



from __future__ import annotations

from itertools import chain
from typing import Any, Callable, Optional, Union

import pyspark.sql.dataframe

from edsnlp import registry
from edsnlp.core.lazy_collection import LazyCollection
from edsnlp.data.base import BaseReader, BaseWriter
from edsnlp.data.converters import (
    FILENAME,
    get_dict2doc_converter,
    get_doc2dict_converter,
    without_filename,
)
from edsnlp.utils.collections import flatten_once


class SparkReader(BaseReader):
    DATA_FIELDS = ("data",)

    def __init__(self, data: pyspark.sql.dataframe.DataFrame):
        import pyspark.sql.dataframe

        self.data = data
        assert isinstance(
            self.data, (pyspark.sql.dataframe.DataFrame, chain)
        ), f"`data` should be a pyspark or koalas DataFrame got {type(data)}"
        super().__init__()

    def read_main(self):
        return ((d, 1) for d in self.data.toLocalIterator())

    def read_worker(self, fragment):
        return [task.asDict(recursive=True) for task in fragment]


@registry.readers.register("spark")
def from_spark(
    data,
    converter: Optional[Union[str, Callable]] = None,
    **kwargs,
) -> LazyCollection:
    """
    The SparkReader (or `edsnlp.data.from_spark`) reads a pyspark (or koalas) DataFrame
    and yields documents. At the moment, only entities and span attributes are loaded.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)
    doc_iterator = edsnlp.data.from_spark(note_df, converter="omop")
    annotated_docs = nlp.pipe(doc_iterator)
    ```

    !!! note "Generator vs list"

        `edsnlp.data.from_spark` returns a
        [LazyCollection][edsnlp.core.lazy_collection.LazyCollection]
        To iterate over the documents multiple times efficiently or to access them by
        index, you must convert it to a list

        ```{ .python .no-check }
        docs = list(edsnlp.data.from_spark(note_df, converter="omop"))
        ```

    Parameters
    ----------
    data: pyspark.sql.dataframe.DataFrame
        The DataFrame to read.
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the rows of the DataFrame to Doc objects.
        These are documented on the [Converters](/data/converters) page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.

    Returns
    -------
    LazyCollection
    """
    data = LazyCollection(reader=SparkReader(data))
    if converter:
        converter, kwargs = get_dict2doc_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)
    return data


class SparkWriter(BaseWriter):
    def __init__(
        self,
        *,
        dtypes: Any = None,
        show_dtypes: bool = True,
    ):
        self.dtypes = dtypes
        self.show_dtypes = show_dtypes

        super().__init__()

    def write_worker(self, records):
        # We flatten in case the converter returns a list
        records = list(flatten_once(records))
        for rec in records:
            rec.pop(FILENAME, None)
        return records, len(records)

    def write_main(self, fragments):
        import pyspark

        from edsnlp.utils.spark_dtypes import (
            schema_warning,
            spark_interpret_dicts_as_rows,
        )

        fragments = map(without_filename, flatten_once(fragments))
        spark = pyspark.sql.SparkSession.builder.enableHiveSupport().getOrCreate()
        rdd = (
            fragments
            if isinstance(fragments, pyspark.RDD)
            else spark.sparkContext.parallelize(fragments)
        )
        with spark_interpret_dicts_as_rows():
            result = spark.createDataFrame(rdd, schema=self.dtypes)

        if self.dtypes is None and self.show_dtypes:
            schema_warning(result.schema)

        return result


@registry.writers.register("spark")
def to_spark(
    data: Union[Any, LazyCollection],
    converter: Optional[Union[str, Callable]] = None,
    dtypes: Any = None,
    show_dtypes: bool = True,
    **kwargs,
):
    """
    `edsnlp.data.to_spark` converts a list of documents into a Spark DataFrame, usually
    one row per document, unless the converter returns a list in which case each entry
    of the resulting list will be stored in its own row.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.covid())

    note_df = sql('''
        select note_id, note_text from note
        where note_text is not null
        limit 500
    ''')

    docs = edsnlp.data.from_spark(note_df, converter="omop")

    docs = nlp.pipe(docs)

    res = edsnlp.data.to_spark(docs, converter="omop")

    res.show()
    ```

    !!! tip "Mac OS X"

        If you are using Mac OS X, you may need to set the following environment
        variable (see [this thread](https://stackoverflow.com/a/52230415)) to run
        pyspark:

        ```{ .python .no-check }
        import os
        os.environ["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"
        ```

    Parameters
    ----------
    data: Union[Any, LazyCollection],
        The data to write (either a list of documents or a LazyCollection).
    dtypes: pyspark.sql.types.StructType
        The schema to use for the DataFrame.
    show_dtypes: bool
        Whether to print the inferred schema (only if `dtypes` is None).
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects before storing
        them in the dataframe. These are documented on the
        [Converters](/data/converters) page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.
    """
    data = LazyCollection.ensure_lazy(data)
    if converter:
        converter, kwargs = get_doc2dict_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)

    return data.write(SparkWriter(dtypes=dtypes, show_dtypes=show_dtypes))



results_diabetes = dict(
    has_match=[
        True,
        True,
        True,
        False,
        True,
        True,
        True,
    ],
    detailled_status=[
        "WITHOUT_COMPLICATION",
        "WITHOUT_COMPLICATION",
        "WITHOUT_COMPLICATION",
        None,
        "WITH_COMPLICATION",
        "WITH_COMPLICATION",
        "WITH_COMPLICATION",
    ],
    assign=None,
    texts=[
        "Présence d'un DT2",
        "Présence d'un DNID",
        "Patient diabétique",
        "Un diabète insipide",
        "Atteinte neurologique d'origine diabétique",
        "Une rétinopathie diabétique",
        "Il y a un mal perforant plantaire",
    ],
)



from typing import List

termination: List[str] = [
    "et",
    "bien que",
    "même si",
    "mais",
    "or",
    "alors que",
    "sauf",
    "cependant",
    "pourtant",
    "cause de",
    "source de",
    "hormis",
    "car",
    "parce que",
    "pourtant",
    "puisque",
    "ni",
    "en raison de",
    "qui",
    "que",
    "ainsi que",
    "avec",
    "toutefois",
    "en dehors",
    "dans le cadre",
    "du fait",
    ".",
    ",",
    ";",
    "...",
    "…",
    "(",
    ")",
    '"',
]



from edsnlp.core import PipelineProtocol
from edsnlp.utils.examples import parse_example

examples = [
    "Patient admis pour <ent code=A01>fièvres typhoïde et paratyphoïde</ent>",
    "Patient admis pour <ent code=C221>C2.21</ent>",
]


def test_cim10(blank_nlp: PipelineProtocol):
    blank_nlp.add_pipe("eds.cim10")

    for text, entities in map(parse_example, examples):
        doc = blank_nlp(text)

        assert len(doc.ents) == len(entities)

        for ent, entity in zip(doc.ents, entities):
            assert ent.text == text[entity.start_char : entity.end_char]
            assert ent.kb_id_ == entity.modifiers[0].value



from edsnlp.utils.regex import make_pattern

from .atomic.delimiters import delimiters

# Pagination
page_patterns = [r"\d\/\d"]

# Phone numbers
phone_patterns = [r"(\d\d" + delimiter + r"){3,}\d\d" for delimiter in delimiters]

false_positive_pattern = make_pattern(page_patterns + phone_patterns)



import re

from edsnlp.pipelines.ner.scores import Score
from edsnlp.utils.examples import parse_example

example = """
CR-REA.
Charlson à l'admission: <ent score_name=charlson score_value=8>8</ent>.
- Charlson:
- IMC: 21
Cette phrase teste un score qui s'appelle TestScore.
La seule valeur admissible est 0.
testScore de 1.
TestScore de <ent score_name=TestScore score_value=0>0</ent>.
Testons également un autre score.
SOFA maximum : <ent score_name=sofa score_value=12 score_method=Maximum>12</ent>.


CR-URG.
PRIORITE: <ent score_name=emergency_priority score_value=2>2</ent>: 2 - Urgence relative.
GEMSA: (<ent score_name=emergency_gemsa score_value=2>2</ent>) Patient non convoque sortant apres consultation
CCMU: Etat clinique jugé stable avec actes diag ou thérapeutiques ( <ent score_name=emergency_ccmu score_value=2>2</ent> )


CONCLUSION

La patiente est atteinte d'un carcinome mammaire infiltrant de type non spécifique, de grade 2 de malignité selon Elston et Ellis
<ent score_name=elston_ellis score_value=2>(architecture : 3 + noyaux : 3 + mitoses : 1)</ent>.

"""  # noqa: E501


def test_scores(blank_nlp):
    blank_nlp.add_pipe(
        "eds.normalizer",
        config=dict(lowercase=True, accents=True, quotes=True, pollution=False),
    )

    def testscore_normalization(raw_score: str):
        if raw_score is not None and int(raw_score) == 0:
            return int(raw_score)

    testscore = Score(
        blank_nlp,
        name="TestScore",
        score_name="TestScore",
        regex=[r"test+score"],
        attr="NORM",
        ignore_excluded=True,
        ignore_space_tokens=False,
        value_extract=r"(\d+)",
        score_normalization=testscore_normalization,
        window=4,
        flags=re.S,
    )

    text, entities = parse_example(example=example)

    blank_nlp.add_pipe("eds.charlson")
    blank_nlp.add_pipe("eds.sofa")
    blank_nlp.add_pipe("eds.elston_ellis")
    blank_nlp.add_pipe("eds.emergency_priority")
    blank_nlp.add_pipe("eds.emergency_ccmu")
    blank_nlp.add_pipe("eds.emergency_gemsa")

    doc = blank_nlp(text)
    doc = testscore(doc)

    for entity, ent in zip(entities, doc.ents):

        for modifier in entity.modifiers:
            assert (
                getattr(ent._, modifier.key) == modifier.value
            ), f"{modifier.key} labels don't match."



from typing import List, Optional, Set, Union

from spacy.tokens import Doc, Span, Token

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.pipes.qualifiers.base import RuleBasedQualifier
from edsnlp.utils.filter import consume_spans, filter_spans
from edsnlp.utils.inclusion import check_inclusion
from edsnlp.utils.resources import get_verbs
from edsnlp.utils.span_getters import SpanGetterArg, get_spans

from . import patterns


def reported_speech_getter(token: Union[Token, Span]) -> Optional[str]:
    if token._.reported_speech is True:
        return "REPORTED"
    elif token._.rspeech is False:
        return "DIRECT"
    else:
        return None


class ReportedSpeechQualifier(RuleBasedQualifier):
    """
    The `eds.reported_speech` component uses a simple rule-based algorithm to detect
    spans that relate to reported speech (eg when the doctor quotes the patient).
    It was designed at AP-HP's EDS.

    Examples
    --------
    The following snippet matches a simple terminology, and checks whether the extracted
    entities are part of a reported speech. It is complete and can be run _as is_.

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    # Dummy matcher
    nlp.add_pipe(eds.matcher(terms=dict(patient="patient", alcool="alcoolisé")))
    nlp.add_pipe(eds.reported_speech())

    text = (
        "Le patient est admis aux urgences ce soir pour une douleur au bras. "
        "Il nie être alcoolisé."
    )

    doc = nlp(text)

    doc.ents
    # Out: (patient, alcoolisé)

    doc.ents[0]._.reported_speech
    # Out: False

    doc.ents[1]._.reported_speech
    # Out: True
    ```

    Extensions
    ----------
    The `eds.reported_speech` component declares two extensions, on both `Span` and
    `Token` objects :

    1. The `reported_speech` attribute is a boolean, set to `True` if the component
       predicts that the span/token is reported.
    2. The `reported_speech_` property is a human-readable string, computed from the
       `reported_speech` attribute. It implements a simple getter function that outputs
       `DIRECT` or `REPORTED`, depending on the value of `reported_speech`.

    Parameters
    ----------
    nlp : PipelineProtocol
        spaCy nlp pipeline to use for matching.
    name : Optional[str]
        The component name.
    quotation : str
        String gathering all quotation cues.
    verbs : List[str]
        List of reported speech verbs.
    following : List[str]
        List of terms following a reported speech.
    preceding : List[str]
        List of terms preceding a reported speech.
    attr : str
        spaCy's attribute to use:
        a string with the value "TEXT" or "NORM",
        or a dict with the key 'term_attr'
        we can also add a key for each regex.
    span_getter : SpanGetterArg
        Which entities should be classified. By default, `doc.ents`
    on_ents_only : Union[bool, str, List[str], Set[str]]
        Whether to look for matches around detected entities only.
        Useful for faster inference in downstream tasks.

        - If True, will look in all ents located in `doc.ents` only
        - If an iterable of string is passed, will additionally look in `doc.spans[key]`
        for each key in the iterable
    within_ents : bool
        Whether to consider cues within entities.
    explain : bool
        Whether to keep track of cues for each entity.

    Authors and citation
    --------------------
    The `eds.reported_speech` component was developed by AP-HP's Data Science team.
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "reported_speech",
        *,
        pseudo: Optional[List[str]] = None,
        preceding: Optional[List[str]] = None,
        following: Optional[List[str]] = None,
        quotation: Optional[List[str]] = None,
        verbs: Optional[List[str]] = None,
        attr: str = "NORM",
        span_getter: Optional[SpanGetterArg] = None,
        on_ents_only: Optional[Union[bool, str, List[str], Set[str]]] = None,
        within_ents: bool = False,
        explain: bool = False,
    ):
        terms = dict(
            pseudo=pseudo or [],
            preceding=patterns.preceding if preceding is None else preceding,
            following=patterns.following if following is None else following,
            quotation=patterns.quotation if quotation is None else quotation,
            verbs=patterns.verbs if verbs is None else verbs,
        )
        terms["verbs"] = self.load_verbs(terms["verbs"])

        quotation = terms.pop("quotation")

        super().__init__(
            nlp=nlp,
            name=name,
            attr=attr,
            explain=explain,
            terms=terms,
            on_ents_only=on_ents_only,
            span_getter=span_getter,
        )

        self.regex_matcher = RegexMatcher(attr=attr)
        self.regex_matcher.build_patterns(dict(quotation=quotation))

        self.within_ents = within_ents
        self.set_extensions()

    def set_extensions(self) -> None:
        super().set_extensions()

        for cls in (Token, Span):
            if not cls.has_extension("reported_speech"):
                cls.set_extension("reported_speech", default=None)

            if not cls.has_extension("reported_speech_"):
                cls.set_extension("reported_speech_", getter=reported_speech_getter)

        if not Span.has_extension("reported_speech_cues"):
            Span.set_extension("reported_speech_cues", default=[])

    def load_verbs(self, verbs: List[str]) -> List[str]:
        """
        Conjugate reporting verbs to specific tenses (trhid person)

        Parameters
        ----------
        verbs: list of reporting verbs to conjugate

        Returns
        -------
        list_rep_verbs: List of reporting verbs conjugated to specific tenses.
        """

        rep_verbs = get_verbs(verbs)

        rep_verbs = rep_verbs.loc[
            (
                (rep_verbs["mode"] == "Indicatif")
                & (rep_verbs["tense"] == "Présent")
                & (rep_verbs["person"].isin(["3s", "3p"]))
            )
            | (rep_verbs["tense"] == "Participe Présent")
            | (rep_verbs["tense"] == "Participe Passé")
        ]

        list_rep_verbs = list(rep_verbs["term"].unique())

        return list_rep_verbs

    def process(self, doc: Doc) -> Doc:
        matches = self.get_matches(doc)
        matches += list(self.regex_matcher(doc, as_spans=True))

        boundaries = self._boundaries(doc)

        # Removes duplicate matches and pseudo-expressions in one statement
        matches = filter_spans(matches, label_to_remove="pseudo")

        entities = list(get_spans(doc, self.span_getter))
        ents = None

        for start, end in boundaries:
            ents, entities = consume_spans(
                entities,
                filter=lambda s: check_inclusion(s, start, end),
                second_chance=ents,
            )

            sub_matches, matches = consume_spans(
                matches, lambda s: start <= s.start < end
            )

            if self.on_ents_only and not ents:
                continue

            sub_preceding = [m for m in sub_matches if m.label_ == "preceding"]
            sub_following = [m for m in sub_matches if m.label_ == "following"]
            sub_verbs = [m for m in sub_matches if m.label_ == "verbs"]
            sub_quotation = [m for m in sub_matches if m.label_ == "quotation"]

            if not self.on_ents_only:
                for token in doc[start:end]:
                    token._.reported_speech = (
                        any(m.end <= token.i for m in sub_preceding + sub_verbs)
                        or any(m.start > token.i for m in sub_following)
                        or any(
                            ((m.start < token.i) & (m.end > token.i + 1))
                            for m in sub_quotation
                        )
                    )
            for ent in ents:
                if self.within_ents:
                    cues = [m for m in sub_preceding + sub_verbs if m.end <= ent.end]
                    cues += [m for m in sub_following if m.start >= ent.start]
                else:
                    cues = [m for m in sub_preceding + sub_verbs if m.end <= ent.start]
                    cues += [m for m in sub_following if m.start >= ent.end]

                cues += [
                    m
                    for m in sub_quotation
                    if (m.start < ent.start) & (m.end > ent.end)
                ]

                reported_speech = bool(cues)
                ent._.reported_speech = ent._.reported_speech or reported_speech

                if self.explain and reported_speech:
                    ent._.reported_speech_cues += cues

                if not self.on_ents_only and reported_speech:
                    for token in ent:
                        token._.reported_speech = True

        return doc



import re
from typing import Any, Dict, List, Union

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.core.contextual_matcher import ContextualMatcher
from edsnlp.utils.deprecation import deprecated_getter_factory
from edsnlp.utils.filter import filter_spans


class DisorderMatcher(ContextualMatcher):
    """
    Base class used to implement various disorders or behaviors extraction pipes

    Parameters
    ----------
    nlp : PipelineProtocol
        spaCy `Language` object.
    name : str
        The name of the pipe
    patterns: Union[Dict[str, Any], List[Dict[str, Any]]]
        The configuration dictionary
    include_assigned : bool
        Whether to include (eventual) assign matches to the final entity
    ignore_excluded : bool
        Whether to skip excluded tokens during matching.
    ignore_space_tokens: bool
        Whether to skip space tokens during matching.
    detailed_status_mapping: Dict[int, str]
        Mapping from integer status (0, 1 or 2) to human-readable string

    alignment_mode : str
        Overwrite alignment mode.
    regex_flags : Union[re.RegexFlag, int]
        RegExp flags to use when matching, filtering and assigning (See
        the [re docs](https://docs.python.org/3/library/re.html#flags))
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: str,
        *,
        label: str,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]],
        include_assigned: bool = True,
        ignore_excluded: bool = True,
        ignore_space_tokens: bool = True,
        detailed_status_mapping: Dict[int, str] = {
            0: "ABSENT",
            1: "PRESENT",
        },
        alignment_mode: str = "expand",
        regex_flags: Union[re.RegexFlag, int] = re.S,
        span_setter: SpanSetterArg,
    ):
        self.detailed_status_mapping = detailed_status_mapping

        super().__init__(
            nlp=nlp,
            name=name,
            label=label,
            attr="NORM",
            patterns=patterns,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
            regex_flags=regex_flags,
            alignment_mode=alignment_mode,
            assign_as_span=True,
            include_assigned=include_assigned,
            span_setter=span_setter,
        )

    def set_extensions(self) -> None:
        super().set_extensions()

        if not Span.has_extension("status"):
            Span.set_extension("status", default=1)
        if not Span.has_extension("negation"):
            Span.set_extension("negation", default=None)
        if not Span.has_extension("detailed_status"):
            Span.set_extension("detailed_status", default="PRESENT")
        if not Span.has_extension("detailled_status"):
            Span.set_extension(
                "detailled_status",
                getter=deprecated_getter_factory("detailed_status", "detailed_status"),
            )
        if not Span.has_extension("negation"):
            Span.set_extension("negation", default=None)
        if not Span.has_extension("negation_"):
            Span.set_extension(
                "negation_",
                getter=lambda item: "NEG"
                if item._.negation is True
                else "AFF"
                if item._.negation is False
                else None,
            )

    def __call__(self, doc: Doc) -> Doc:
        """
        Tags entities.

        Parameters
        ----------
        doc : Doc
            spaCy Doc object

        Returns
        -------
        doc : Doc
            annotated spaCy Doc object
        """
        spans = list(self.process(doc))
        for span in spans:
            span._.detailed_status = self.detailed_status_mapping[span._.status]
            if span._.status == 0:
                span._.negation = True

        self.set_spans(doc, filter_spans(spans))

        return doc



import re
from typing import Any, Callable, Dict, List, Optional, Union

from edsnlp.core import PipelineProtocol, registry
from edsnlp.pipes.base import SpanSetterArg

from .patterns import regex, score_normalization_str, value_extract
from .sofa import SofaMatcher

DEFAULT_CONFIG = dict(
    regex=regex,
    value_extract=value_extract,
    score_normalization=score_normalization_str,
    attr="NORM",
    window=10,
    ignore_excluded=False,
    ignore_space_tokens=False,
    flags=0,
    label="sofa",
    span_setter={"ents": True, "sofa": True},
)


@registry.factory.register(
    "eds.sofa",
    assigns=["doc.ents", "doc.spans"],
    deprecated=[
        "SOFA",
        "eds.SOFA",
    ],
)
def create_component(
    nlp: PipelineProtocol,
    name: Optional[str] = "sofa",
    *,
    regex: List[str] = regex,
    value_extract: List[Dict[str, str]] = value_extract,
    score_normalization: Union[
        str, Callable[[Union[str, None]], Any]
    ] = score_normalization_str,
    attr: str = "NORM",
    window: int = 10,
    ignore_excluded: bool = False,
    ignore_space_tokens: bool = False,
    flags: Union[re.RegexFlag, int] = 0,
    label: str = "sofa",
    span_setter: SpanSetterArg = {"ents": True, "sofa": True},
):
    '''
    The `eds.sofa` component extracts
    [Sequential Organ Failure Assessment (SOFA) scores](\
    https://www.mdcalc.com/calc/691/sequential-organ-failure-assessment-sofa-score),
    used to track a person's status during the stay in an intensive care unit to
    determine the extent of a person's organ function or rate failure.

    Examples
    --------

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.sofa())

    text = """
    SOFA (à 24H) : 12.
    OMS:
    """

    doc = nlp(text)
    doc.ents
    # Out: (SOFA (à 24H) : 12,)
    ```

    Extensions
    ----------
    Each extraction exposes 3 extensions:

    ```python
    ent = doc.ents[0]

    ent._.score_name
    # Out: 'sofa'

    ent._.score_value
    # Out: 12

    ent._.score_method
    # Out: '24H'
    ```

    Score method can here be "24H", "Maximum", "A l'admission" or "Non précisée"

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : Optional[str]
        The name of the component
    regex : List[str]
        A list of regexes to identify the SOFA score
    attr : str
        Whether to match on the text ('TEXT') or on the normalized text ('CUSTOM_NORM')
    value_extract : Dict[str, str]
        Regex to extract the score value
    score_normalization : Callable[[Union[str,None]], Any]
        Function that takes the "raw" value extracted from the `value_extract` regex,
        and should return
        - None if no score could be extracted
        - The desired score value else
    window : int
        Number of token to include after the score's mention to find the
        score's value
    ignore_excluded : bool
        Whether to ignore excluded spans
    ignore_space_tokens : bool
        Whether to ignore space tokens
    flags : Union[re.RegexFlag, int]
        Flags to pass to the regex
    label: str
        Label name to use for the `Span` object and the extension
    span_setter: SpanSetterArg
        How to set matches on the doc
    '''
    return SofaMatcher(
        nlp,
        name=name,
        regex=regex,
        value_extract=value_extract,
        score_normalization=score_normalization,
        attr=attr,
        window=window,
        ignore_excluded=ignore_excluded,
        ignore_space_tokens=ignore_space_tokens,
        flags=flags,
        label=label,
        span_setter=span_setter,
    )



import pytest

from edsnlp.core import PipelineProtocol
from edsnlp.utils.examples import parse_example

example = "1g de <ent kb_id=paracetamol>doliprane</ent>"


@pytest.mark.parametrize("term_matcher", ["exact", "simstring"])
def test_terminology(blank_nlp: PipelineProtocol, term_matcher: str):
    blank_nlp.add_pipe(
        "eds.terminology",
        config=dict(
            label="drugs",
            terms=dict(paracetamol=["doliprane", "tylenol", "paracetamol"]),
            attr="NORM",
            term_matcher=term_matcher,
        ),
    )

    text, entities = parse_example(example)

    doc = blank_nlp(text)

    assert len(entities) == len(doc.ents)

    for ent, entity in zip(doc.ents, entities):
        assert ent.text == text[entity.start_char : entity.end_char]
        assert ent.kb_id_ == entity.modifiers[0].value



import importlib
import subprocess
import sys

import pytest
import torch

import edsnlp
from edsnlp.utils.package import package


def test_blank_package(nlp, tmp_path):
    # Missing metadata makes poetry fail due to missing author / description
    if not isinstance(nlp, edsnlp.Pipeline):
        pytest.skip("Only running for edsnlp.Pipeline")

    with pytest.raises(Exception):
        package(
            pipeline=nlp,
            root_dir=tmp_path,
            name="test-model-fail",
            metadata={},
            project_type="poetry",
        )

    nlp.package(
        root_dir=tmp_path,
        name="test-model",
        metadata={
            "description": "A test model",
            "authors": "Test Author <test.author@mail.com>",
        },
        project_type="poetry",
        distributions=["wheel"],
    )
    assert (tmp_path / "dist").is_dir()
    assert (tmp_path / "build").is_dir()
    assert (tmp_path / "dist" / "test_model-0.1.0-py3-none-any.whl").is_file()
    assert not (tmp_path / "dist" / "test_model-0.1.0.tar.gz").is_file()
    assert (tmp_path / "build" / "test-model").is_dir()


@pytest.mark.parametrize("package_name", ["my-test-model", None])
def test_package_with_files(nlp, tmp_path, package_name):
    if not isinstance(nlp, edsnlp.Pipeline):
        pytest.skip("Only running for edsnlp.Pipeline")

    nlp.to_disk(tmp_path / "model")

    ((tmp_path / "test_model").mkdir(parents=True))
    (tmp_path / "test_model" / "__init__.py").write_text(
        """\
print("Hello World!")
"""
    )
    (tmp_path / "pyproject.toml").write_text(
        """\
[build-system]
requires = ["poetry-core>=1.0.0"]
build-backend = "poetry.core.masonry.api"

[tool.poetry]
name = "test-model"
version = "0.0.0"
description = "A test model"
authors = ["Test Author <test.author@mail.com>"]

[tool.poetry.dependencies]
python = "^3.7"
torch = "^{}"
""".format(torch.__version__.split("+")[0])
    )

    with pytest.raises(ValueError):
        package(
            pipeline=nlp,
            root_dir=tmp_path,
            version="0.1.0",
            name=package_name,
            metadata={
                "description": "Wrong description",
                "authors": "Test Author <test.author@mail.com>",
            },
        )

    package(
        name=package_name,
        pipeline=tmp_path / "model",
        root_dir=tmp_path,
        check_dependencies=True,
        version="0.1.0",
        distributions=None,
        metadata={
            "description": "A test model",
            "authors": "Test Author <test.author@mail.com>",
        },
    )

    module_name = "test_model" if package_name is None else "my_test_model"

    assert (tmp_path / "dist").is_dir()
    assert (tmp_path / "dist" / f"{module_name}-0.1.0.tar.gz").is_file()
    assert (tmp_path / "dist" / f"{module_name}-0.1.0-py3-none-any.whl").is_file()
    assert (tmp_path / "pyproject.toml").is_file()

    # pip install the whl file
    print(
        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                str(tmp_path / "dist" / f"{module_name}-0.1.0-py3-none-any.whl"),
                "--force-reinstall",
            ],
            stderr=subprocess.STDOUT,
        )
    )

    module = importlib.import_module(module_name)

    assert module.__version__ == "0.1.0"

    with open(module.__file__) as f:
        assert f.read() == (
            (
                """\
print("Hello World!")
"""
                if package_name is None
                else ""
            )
            + f"""
# -----------------------------------------
# This section was autogenerated by edsnlp
# -----------------------------------------

import edsnlp
from pathlib import Path
from typing import Optional, Dict, Any

__version__ = '{module.__version__}'

def load(
    overrides: Optional[Dict[str, Any]] = None,
) -> edsnlp.Pipeline:
    artifacts_path = Path(__file__).parent / "artifacts"
    model = edsnlp.load(artifacts_path, overrides=overrides)
    return model
"""
        )
    module.load()
    edsnlp.load(module_name)


@pytest.fixture(scope="session", autouse=True)
def clean_after():
    yield

    print(
        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "pip",
                "uninstall",
                "-y",
                "test-model",
            ],
            stderr=subprocess.STDOUT,
        )
    )

    print(
        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "pip",
                "uninstall",
                "-y",
                "my-test-model",
            ],
            stderr=subprocess.STDOUT,
        )
    )



from pytest import fixture
from spacy.tokens import Span

import edsnlp
from edsnlp.pipes.trainable.embeddings.transformer.transformer import Transformer
from edsnlp.pipes.trainable.span_qualifier.span_qualifier import (
    TrainableSpanQualifier,
)

if not Span.has_extension("label"):
    Span.set_extension("label", default=None)

if not Span.has_extension("event_type"):
    Span.set_extension("event_type", default=None)

if not Span.has_extension("test_negated"):
    Span.set_extension("test_negated", default=False)


@fixture
def gold():
    blank_nlp = edsnlp.blank("eds")
    doc1 = blank_nlp.make_doc("Arret du ttt si folfox inefficace. Une autre phrase.")

    doc1.spans["sc"] = [
        Span(doc1, 4, 5, "drug"),  # "folfox"
        Span(doc1, 0, 1, "event"),  # "Arret"
        Span(doc1, 3, 4, "criteria"),  # "si"
    ]
    doc1.spans["sc"][0]._.test_negated = False
    doc1.spans["sc"][1]._.test_negated = True
    doc1.spans["sc"][2]._.test_negated = False
    doc1.spans["sc"][1]._.event_type = "stop"
    doc1.spans["to_embed"] = [doc1[0:5], doc1[7:11]]

    doc1.spans["sent"] = [Span(doc1, 0, 6, "sent")]

    return [doc1]


def test_span_getter(gold):
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        "eds.transformer",
        name="transformer",
        config=dict(
            model="prajjwal1/bert-tiny", window=128, stride=96, span_getter="to_embed"
        ),
    )
    nlp.add_pipe(
        "eds.span_qualifier",
        name="qualifier",
        config={
            "embedding": {
                "@factory": "eds.span_pooler",
                "embedding": nlp.get_pipe("transformer"),
                "span_getter": ["ents", "sc"],
            },
            "qualifiers": ["_.test_negated", "_.event_type"],
        },
    )
    trf: Transformer = nlp.get_pipe("transformer")
    qlf: TrainableSpanQualifier = nlp.get_pipe("qualifier")
    qlf.post_init(gold, set())
    batch = qlf.prepare_batch([doc.copy() for doc in gold], supervision=True)
    input_ids = batch["embedding"]["embedding"]["input_ids"]
    mask = batch["embedding"]["embedding"]["input_mask"]
    tok = trf.tokenizer
    assert len(input_ids) == 2
    assert tok.decode(input_ids[0][mask[0]]) == "[CLS] arret du ttt si folfox [SEP]"
    assert tok.decode(input_ids[1][mask[1]]) == "[CLS] une autre phrase. [SEP]"



import pytest
import spacy
from spacy.tokens import Span
from spacy.tokens.underscore import Underscore

import edsnlp


def test_warn_value_extension():
    old_value_extension = Underscore.span_extensions.pop("value", None)
    try:
        Underscore._extensions = {}
        Span.set_extension("value", getter=lambda span: "stuff")
        existing_nlp = spacy.blank("fr")
        with pytest.warns(UserWarning) as record:
            existing_nlp.add_pipe(
                "eds.terminology",
                name="test",
                config=dict(label="Any", terms={}),
            )

        assert any(
            "A Span extension 'value' already exists with a different getter"
            in str(r.message)
            for r in record
        )
    finally:
        Underscore.span_extensions.pop("value", None)
        if old_value_extension is not None:
            Underscore.span_extensions["value"] = old_value_extension


def test_value_extension():
    # From https://github.com/aphp/edsnlp/issues/220

    # Setting up a first pipeline
    existing_nlp = spacy.blank("fr")
    existing_nlp.add_pipe(
        "eds.terminology",
        name="test",
        config=dict(label="Any", terms={}),
    )

    # Setting up another custom pipeline somewhere else in the code
    nlp = edsnlp.blank("eds")
    text = "hello this is a test"
    doc = nlp(text)
    my_span = doc[0:3]
    my_span._.value = "CustomValue"

    assert my_span._.value == "CustomValue"



from typing import Generic, Optional

import torch
from typing_extensions import TypedDict

from edsnlp.core.torch_component import BatchInput, TorchComponent  # noqa: F401
from edsnlp.utils.span_getters import SpanGetter

WordEmbeddingBatchOutput = TypedDict(
    "WordEmbeddingBatchOutput",
    {
        "embeddings": torch.Tensor,
        "mask": torch.Tensor,
    },
)


class WordEmbeddingComponent(
    Generic[BatchInput], TorchComponent[WordEmbeddingBatchOutput, BatchInput]
):
    span_getter: Optional[SpanGetter]
    output_size: int


SpanEmbeddingBatchOutput = TypedDict(
    "SpanEmbeddingBatchOutput",
    {
        "embeddings": torch.Tensor,
        "mask": torch.Tensor,
    },
)


class SpanEmbeddingComponent(
    Generic[BatchInput], TorchComponent[SpanEmbeddingBatchOutput, BatchInput]
):
    span_getter: Optional[SpanGetter]
    output_size: int



from edsnlp.utils.regex import make_pattern

letter_months = [
    r"(?P<month_01>janvier|janv\.?)",
    r"(?P<month_02>f[ée]vrier|f[ée]v\.?)",
    r"(?P<month_03>mars|mar\.?)",
    r"(?P<month_04>avril|avr\.?)",
    r"(?P<month_05>mai)",
    r"(?P<month_06>juin)",
    r"(?P<month_07>juillet|juill?\.?)",
    r"(?P<month_08>ao[uû]t)",
    r"(?P<month_09>septembre|sept?\.?)",
    r"(?P<month_10>octobre|oct\.?)",
    r"(?P<month_11>novembre|nov\.?)",
    r"(?P<month_12>d[ée]cembre|d[ée]c\.?)",
]


letter_month_pattern = make_pattern(letter_months, with_breaks=True)

numeric_month_pattern = r"(?<!\d)(0?[1-9]|1[0-2])(?!\d)"
lz_numeric_month_pattern = r"(?<!\d)(0[1-9]|1[0-2])(?!\d)"

numeric_month_pattern = f"(?P<month>{numeric_month_pattern})"
lz_numeric_month_pattern = f"(?P<month>{lz_numeric_month_pattern})"
month_pattern = f"({letter_month_pattern}|{numeric_month_pattern})"



import io
import os
import re
import shutil
import subprocess
import sys
from contextlib import contextmanager
from pathlib import Path
from types import FunctionType, ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import build
import dill
import toml
from build.__main__ import build_package, build_package_via_sdist
from confit import Cli
from dill._dill import save_function as dill_save_function
from dill._dill import save_module as dill_save_module
from dill._dill import save_type as dill_save_type

try:
    import importlib_metadata
except ImportError:  # pragma: no cover
    import importlib.metadata as importlib_metadata
from loguru import logger
from typing_extensions import Literal

import edsnlp

py_version = f"{sys.version_info.major}.{sys.version_info.minor}"


def get_package(obj: Type):
    # Retrieve the __package__ attribute of the module of a type, if possible.
    # And returns the package version as well
    try:
        if isinstance(obj, ModuleType):
            module_name = obj.__name__
        else:
            module_name = obj.__module__
        if module_name == "__main__":
            raise Exception(f"Could not find package of {obj}")
        module = __import__(module_name, fromlist=["__package__"])
        package = module.__package__.split(".")[0]
        try:
            version = importlib_metadata.version(package)
        except (importlib_metadata.PackageNotFoundError, ValueError):
            return None
        return package, version
    except (ImportError, AttributeError):
        raise Exception(f"Cound not find package of type {obj}")


def save_type(pickler, obj, *args, **kwargs):
    package_name = get_package(obj)
    if package_name is not None:
        pickler.packages.add(package_name)
    dill_save_type(pickler, obj, *args, **kwargs)


def save_function(pickler, obj, *args, **kwargs):
    package_name = get_package(obj)
    if package_name is not None:
        pickler.packages.add(package_name)
    return dill_save_function(pickler, obj, *args, **kwargs)


def save_module(pickler, obj, *args, **kwargs):
    package_name = get_package(obj)
    if package_name is not None:
        pickler.packages.add(package_name)
    return dill_save_module(pickler, obj, *args, **kwargs)


class PackagingPickler(dill.Pickler):
    dispatch = dill.Pickler.dispatch.copy()

    dispatch[FunctionType] = save_function
    dispatch[type] = save_type
    dispatch[ModuleType] = save_module

    def __init__(self, *args, **kwargs):
        self.file = io.BytesIO()
        super().__init__(self.file, *args, **kwargs)
        self.packages = set()


def get_deep_dependencies(obj):
    pickler = PackagingPickler(byref=True)
    pickler.dump(obj)
    return sorted(pickler.packages)


app = Cli(pretty_exceptions_show_locals=False, pretty_exceptions_enable=False)


def snake_case(s):
    # From https://www.w3resource.com/python-exercises/string/python-data-type-string-exercise-97.php  # noqa E501
    return "_".join(
        re.sub(
            "([A-Z][a-z]+)", r" \1", re.sub("([A-Z]+)", r" \1", s.replace("-", " "))
        ).split()
    ).lower()


class ModuleName(str):
    def __new__(cls, *args, **kwargs):
        raise NotImplementedError("ModuleName is only meant for typing.")

    @classmethod
    def __get_validators__(self):
        yield self.validate

    @classmethod
    def validate(cls, value, config=None):
        if not isinstance(value, str):
            raise TypeError("string required")

        if not re.match(
            r"^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$", value, flags=re.IGNORECASE
        ):
            raise ValueError("invalid identifier")
        return value


if TYPE_CHECKING:
    ModuleName = str  # noqa F811

POETRY_SNIPPET = """\
from poetry.core.masonry.builders.sdist import SdistBuilder
from poetry.factory import Factory
from poetry.core.masonry.utils.module import ModuleOrPackageNotFound
import sys
# Initialize the Poetry object for the current project
poetry = Factory().create_poetry("__root_dir__")

# Initialize the builder
try:
    builder = SdistBuilder(poetry, None, None)
    # Get the list of files to include
    files = builder.find_files_to_add()
except ModuleOrPackageNotFound:
    if not poetry.package.packages:
        print([])
        sys.exit(0)

print([
    {k: v for k, v in {
    "include": getattr(include, '_include'),
    "from": getattr(include, 'source', None),
    "formats": getattr(include, 'formats', None),
    }.items() if v}
    for include in builder._module.includes
])


# Print the list of files
for file in files:
    print(file.path)
"""

INIT_PY = """
# -----------------------------------------
# This section was autogenerated by edsnlp
# -----------------------------------------

import edsnlp
from pathlib import Path
from typing import Optional, Dict, Any

__version__ = {__version__}

def load(
    overrides: Optional[Dict[str, Any]] = None,
) -> edsnlp.Pipeline:
    artifacts_path = Path(__file__).parent / "{artifacts_dir}"
    model = edsnlp.load(artifacts_path, overrides=overrides)
    return model
"""


# def parse_authors_as_dicts(authors):
#     authors = [authors] if isinstance(authors, str) else authors
#     return [
#         dict(zip(("name", "email"), re.match(r"(.*) <(.*)>", author).groups()))
#         if isinstance(author, str)
#         else author
#         for author in authors
#     ]


def parse_authors_as_strings(authors):
    authors = [authors] if isinstance(authors, str) else authors
    return [
        author if isinstance(author, str) else f"{author['name']} <{author['email']}>"
        for author in authors
    ]


class PoetryPackager:
    def __init__(
        self,
        pyproject: Optional[Dict[str, Any]],
        pipeline: Union[Path, "edsnlp.Pipeline"],
        version: Optional[str],
        name: Optional[ModuleName],
        root_dir: Path = ".",
        build_dir: Path = "build",
        dist_dir: Path = "dist",
        artifacts_name: ModuleName = "artifacts",
        dependencies: Optional[Sequence[Tuple[str, str]]] = None,
        metadata: Optional[Dict[str, Any]] = {},
    ):
        self.poetry_bin_path = (
            subprocess.run(["which", "poetry"], stdout=subprocess.PIPE)
            .stdout.decode()
            .strip()
        )
        self.version = version
        self.name = name
        self.pyproject = pyproject
        self.root_dir = root_dir.resolve()
        self.dependencies = dependencies
        self.pipeline = pipeline
        self.artifacts_name = artifacts_name
        self.dist_dir = (
            dist_dir if Path(dist_dir).is_absolute() else self.root_dir / dist_dir
        )

        with self.ensure_pyproject(metadata):
            python_executable = (
                Path(self.poetry_bin_path).read_text().split("\n")[0][2:]
            )
            result = subprocess.run(
                [
                    *python_executable.split(),
                    "-c",
                    POETRY_SNIPPET.replace("__root_dir__", str(self.root_dir)),
                ],
                stdout=subprocess.PIPE,
                cwd=self.root_dir,
            )
            if result.returncode != 0:
                raise Exception()
            out = result.stdout.decode().strip().split("\n")

        self.poetry_packages = eval(out[0])
        self.build_dir = (
            build_dir if Path(build_dir).is_absolute() else root_dir / build_dir
        ) / self.name
        self.file_paths = [self.root_dir / file_path for file_path in out[1:]]

        logger.info(f"root_dir: {self.root_dir}")
        logger.info(f"build_dir: {self.build_dir}")
        logger.info(f"artifacts_name: {self.artifacts_name}")
        logger.info(f"name: {self.name}")

    @contextmanager
    def ensure_pyproject(self, metadata):
        """Generates a Poetry based pyproject.toml"""
        metadata = dict(metadata)
        new_pyproject = self.pyproject is None
        if "authors" in metadata:
            metadata["authors"] = parse_authors_as_strings(metadata["authors"])
        try:
            if new_pyproject:
                self.pyproject = {
                    "build-system": {
                        "requires": ["poetry-core>=1.0.0"],
                        "build-backend": "poetry.core.masonry.api",
                    },
                    "tool": {
                        "poetry": {
                            **metadata,
                            "name": self.name,
                            "version": self.version or "0.1.0",
                            "dependencies": {
                                "python": f">={py_version},<4.0",
                                **{
                                    dep_name: f"^{dep_version}"
                                    for dep_name, dep_version in self.dependencies
                                },
                            },
                        },
                    },
                }
                (self.root_dir / "pyproject.toml").write_text(
                    toml.dumps(self.pyproject)
                )
            else:
                self.name = (
                    self.pyproject["tool"]["poetry"]["name"]
                    if self.name is None
                    else self.name
                )
                for key, value in metadata.items():
                    pyproject_value = self.pyproject["tool"]["poetry"].get(key)
                    if pyproject_value != metadata[key]:
                        raise ValueError(
                            f"Field {key} in pyproject.toml doesn't match the one "
                            f"passed as argument, you should remove it from the "
                            f"metadata parameter. Avoid using metadata if you already "
                            f"have a pyproject.toml file.\n"
                            f"pyproject.toml:\n {pyproject_value}\n"
                            f"metadata:\n {value}"
                        )
            yield
        except Exception:
            if new_pyproject:
                os.remove(self.root_dir / "pyproject.toml")
            raise

    def list_files_to_add(self):
        # Extract python from the shebang in the poetry executable
        return self.file_paths

    def build(
        self,
        distributions: Sequence[str] = (),
        config_settings: Optional[build.ConfigSettingsType] = None,
        isolation: bool = True,
        skip_dependency_check: bool = False,
    ):
        logger.info(f"Building package {self.name}")

        if distributions:
            build_call = build_package
        else:
            build_call = build_package_via_sdist
            distributions = ["wheel"]
        build_call(
            srcdir=self.build_dir,
            outdir=self.dist_dir,
            distributions=distributions,
            config_settings=config_settings,
            isolation=isolation,
            skip_dependency_check=skip_dependency_check,
        )

    def update_pyproject(self):
        # Replacing project name
        old_name = self.pyproject["tool"]["poetry"]["name"]
        self.pyproject["tool"]["poetry"]["name"] = self.name
        logger.info(
            f"Replaced project name {old_name!r} with {self.name!r} in poetry based "
            f"project"
        )

        if self.version is not None:
            old_version = self.pyproject["tool"]["poetry"]["version"]
            self.pyproject["tool"]["poetry"]["version"] = self.version
            logger.info(
                f"Replaced project version {old_version!r} with {self.version!r} in "
                f"poetry based project"
            )

        # Adding artifacts to include in pyproject.toml
        snake_name = snake_case(self.name.lower())
        included = self.pyproject["tool"]["poetry"].setdefault("include", [])
        included.append(f"{snake_name}/{self.artifacts_name}/**")

        packages = list(self.poetry_packages)
        packages.append({"include": snake_name})
        self.pyproject["tool"]["poetry"]["packages"] = packages

    def make_src_dir(self):
        snake_name = snake_case(self.name.lower())
        package_dir = self.build_dir / snake_name
        shutil.rmtree(package_dir, ignore_errors=True)
        os.makedirs(package_dir, exist_ok=True)
        build_artifacts_dir = package_dir / self.artifacts_name
        for file_path in self.list_files_to_add():
            new_file_path = self.build_dir / Path(file_path).relative_to(self.root_dir)
            if isinstance(self.pipeline, Path) and self.pipeline in file_path.parents:
                raise Exception(
                    f"Pipeline ({self.artifacts_name}) is already "
                    "included in the package's data, you should "
                    "remove it from the pyproject.toml metadata."
                )
            os.makedirs(new_file_path.parent, exist_ok=True)
            logger.info(f"COPY {file_path}" f" TO {new_file_path}")
            shutil.copy(file_path, new_file_path)

        self.update_pyproject()

        # Write pyproject.toml
        (self.build_dir / "pyproject.toml").write_text(toml.dumps(self.pyproject))

        if isinstance(self.pipeline, Path):
            # self.pipeline = edsnlp.load(self.pipeline)
            shutil.copytree(
                self.pipeline,
                build_artifacts_dir,
            )
        else:
            self.pipeline.to_disk(build_artifacts_dir)
        os.makedirs(package_dir, exist_ok=True)
        with open(package_dir / "__init__.py", mode="a") as f:
            f.write(
                INIT_PY.format(
                    __version__=repr(self.version),
                    artifacts_dir=os.path.relpath(build_artifacts_dir, package_dir),
                )
            )


@app.command(name="package")
def package(
    pipeline: Union[Path, "edsnlp.Pipeline"],
    name: Optional[ModuleName] = None,
    root_dir: Path = ".",
    build_dir: Path = "build",
    dist_dir: Path = "dist",
    artifacts_name: ModuleName = "artifacts",
    check_dependencies: bool = False,
    project_type: Optional[Literal["poetry", "setuptools"]] = None,
    version: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = {},
    distributions: Optional[Sequence[Literal["wheel", "sdist"]]] = ["wheel"],
    config_settings: Optional[Mapping[str, Union[str, Sequence[str]]]] = None,
    isolation: bool = True,
    skip_build_dependency_check: bool = False,
):
    # root_dir = Path(".").resolve()
    pyproject_path = root_dir / "pyproject.toml"

    if not pyproject_path.exists():
        check_dependencies = True
        if name is None:
            raise ValueError(
                f"No pyproject.toml could be found in the root directory {root_dir}, "
                f"you need to create one, or fill the name parameter."
            )

    dependencies = None
    if check_dependencies:
        if isinstance(pipeline, Path):
            pipeline = edsnlp.load(pipeline)
        dependencies = get_deep_dependencies(pipeline)
        for dep in dependencies:
            print("DEPENDENCY", dep[0].ljust(30), dep[1])

    root_dir = root_dir.resolve()

    pyproject = None
    if pyproject_path.exists():
        pyproject = toml.loads((root_dir / "pyproject.toml").read_text())

        if "tool" in pyproject and "poetry" in pyproject["tool"]:
            project_type = "poetry"

    if project_type == "poetry":
        packager = PoetryPackager(
            pyproject=pyproject,
            pipeline=pipeline,
            name=name,
            version=version,
            root_dir=root_dir,
            build_dir=build_dir,
            dist_dir=dist_dir,
            artifacts_name=artifacts_name,
            dependencies=dependencies,
            metadata=metadata,
        )
    else:
        raise Exception(
            "Could not infer project type, only poetry based projects are "
            "supported for now"
        )

    packager.make_src_dir()
    packager.build(
        distributions=distributions,
        config_settings=config_settings,
        isolation=isolation,
        skip_dependency_check=skip_build_dependency_check,
    )



import fnmatch
import os
from pathlib import Path
from typing import Optional, Tuple, Union

import pyarrow.fs
from fsspec import AbstractFileSystem
from fsspec import __version__ as fsspec_version
from fsspec.implementations.arrow import ArrowFSWrapper

FileSystem = Union[AbstractFileSystem, pyarrow.fs.FileSystem]

if fsspec_version < "2023.3.0":
    # Ugly hack to make fsspec's arrow implementation work in python 3.7
    # since arrow requires files to be seekable, and the default fsspec
    # open(..., seekable) parameter is False
    # See https://github.com/fsspec/filesystem_spec/pull/1186
    ArrowFSWrapper._open.__wrapped__.__defaults__ = ("rb", None, True)


def walk_match(
    fs: FileSystem,
    root: str,
    file_pattern: str,
    recursive: bool = True,
) -> list:
    if fsspec_version >= "2023.10.0":
        # Version fixes fsspec glob https://github.com/fsspec/filesystem_spec/pull/1329
        glob_str = os.path.join(root, "**" if recursive else "", file_pattern)
        return fs.glob(glob_str)
    return [
        os.path.join(dirpath, f)
        for dirpath, dirnames, files in fs.walk(
            root,
            maxdepth=None if recursive else 1,
        )
        for f in fnmatch.filter(files, file_pattern)
    ]


def normalize_fs_path(
    filesystem: Optional[FileSystem],
    path: Union[str, Path],
) -> Tuple[AbstractFileSystem, str]:
    path = str(path)

    if filesystem is None or (isinstance(path, str) and "://" in path):
        path = (
            os.path.abspath(path)
            if isinstance(path, Path) or "://" in path
            else f"file://{os.path.abspath(path)}"
        )
        inferred_fs, fs_path = pyarrow.fs.FileSystem.from_uri(path)
        filesystem = filesystem or inferred_fs
        assert inferred_fs.type_name == filesystem.type_name, (
            f"Protocol {inferred_fs.type_name} in path does not match "
            f"filesystem {filesystem.type_name}"
        )
        path = fs_path

    return (
        ArrowFSWrapper(filesystem)
        if isinstance(filesystem, pyarrow.fs.FileSystem)
        else filesystem
    ), path



import os
import pickle
import tempfile
from collections import defaultdict
from enum import Enum
from functools import lru_cache
from math import sqrt
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Union

import pysimstring.simstring as simstring
from spacy.tokens import Doc, Span
from spacy.vocab import Vocab
from tqdm import tqdm

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.utils import ATTRIBUTES, get_text, normalize_token_attr


class SimstringWriter:
    def __init__(self, path: Union[str, Path]):
        """
        A context class to write a simstring database

        Parameters
        ----------
        path: Union[str, Path]
            Path to database
        """
        os.makedirs(path, exist_ok=True)
        self.path = path

    def __enter__(self):
        path = os.path.join(self.path, "terms.simstring")
        self.db = simstring.writer(path, 3, False, True)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.db.close()

    def insert(self, term):
        self.db.insert(term)


class SimilarityMeasure(str, Enum):
    jaccard = "jaccard"
    dice = "dice"
    overlap = "overlap"
    cosine = "cosine"


class SimstringMatcher:
    def __init__(
        self,
        vocab: Vocab,
        path: Optional[Union[Path, str]] = None,
        measure: SimilarityMeasure = SimilarityMeasure.dice,
        threshold: float = 0.75,
        windows: int = 5,
        ignore_excluded: bool = False,
        ignore_space_tokens: bool = False,
        attr: str = "NORM",
    ):
        """
        PhraseMatcher that allows to skip excluded tokens.
        Heavily inspired by https://github.com/Georgetown-IR-Lab/QuickUMLS

        Parameters
        ----------
        vocab : Vocab
            spaCy vocabulary to match on.
        path: Optional[Union[Path, str]]
            Path where we will store the precomputed patterns
        measure: SimilarityMeasure
            Name of the similarity measure.
            One of [jaccard, dice, overlap, cosine]
        windows: int
            Maximum number of words in a candidate span
        threshold: float
            Minimum similarity value to match a concept's synonym
        ignore_excluded : Optional[bool]
            Whether to exclude tokens that have an EXCLUDED tag, by default False
        ignore_space_tokens : Optional[bool]
            Whether to exclude tokens that have a "SPACE" tag, by default False
        attr : str
            Default attribute to match on, by default "TEXT".
            Can be overridden in the `add` method.
            To match on a custom attribute, prepend the attribute name with `_`.
        """

        assert measure in (
            SimilarityMeasure.jaccard,
            SimilarityMeasure.dice,
            SimilarityMeasure.overlap,
            SimilarityMeasure.cosine,
        )

        self.vocab = vocab
        self.windows = windows
        self.measure = measure
        self.threshold = threshold
        self.ignore_excluded = ignore_excluded
        self.ignore_space_tokens = ignore_space_tokens
        self.attr = attr

        if path is None:
            path = tempfile.mkdtemp()
        self.path = Path(path)

        self.ss_reader = None
        self.syn2cuis = None

    def build_patterns(
        self,
        nlp: PipelineProtocol,
        terms: Dict[str, Iterable[str]],
        progress: bool = False,
    ):
        """
        Build patterns and adds them for matching.

        Parameters
        ----------
        nlp : PipelineProtocol
            The instance of the spaCy language class.
        terms : Patterns
            Dictionary of label/terms, or label/dictionary of terms/attribute.
        progress: bool
            Whether to track progress when preprocessing terms
        """

        self.ss_reader = None
        self.syn2cuis = None

        syn2cuis = defaultdict(lambda: [])

        matched_attr = normalize_token_attr(self.attr)
        assert matched_attr is not None, "Unsupported attribute for matching"
        token_pipelines = [
            name
            for name, pipe in nlp.pipeline
            if name not in nlp.disabled
            and matched_attr
            in {normalize_token_attr(ass) for ass in nlp.get_pipe_meta(name).assigns}
        ]
        with nlp.select_pipes(enable=token_pipelines):
            with SimstringWriter(self.path) as ss_db:
                for cui, synset in tqdm(terms.items()) if progress else terms.items():
                    for term in nlp.pipe(synset):
                        norm_text = get_text(
                            term,
                            self.attr,
                            ignore_excluded=self.ignore_excluded,
                            ignore_space_tokens=self.ignore_space_tokens,
                        )
                        term = "##" + norm_text + "##"
                        ss_db.insert(term)
                        syn2cuis[term].append(cui)
        syn2cuis = {term: tuple(sorted(set(cuis))) for term, cuis in syn2cuis.items()}
        with open(self.path / "cui-db.pkl", "wb") as f:
            pickle.dump(syn2cuis, f)

    def load(self):
        if self.ss_reader is None:
            self.ss_reader = simstring.reader(
                os.path.join(self.path, "terms.simstring")
            )
            self.ss_reader.measure = getattr(simstring, self.measure)
            self.ss_reader.threshold = self.threshold

            with open(os.path.join(self.path, "cui-db.pkl"), "rb") as f:
                self.syn2cuis = pickle.load(f)

    def __call__(self, doc, as_spans=False):
        self.load()

        root = getattr(doc, "doc", doc)
        if root.has_annotation("IS_SENT_START"):
            sents = tuple(doc.sents)
        else:
            sents = (doc,)

        ents: List[Tuple[str, int, int, float]] = []

        for sent in sents:
            text, offsets = get_text_and_offsets(
                doclike=sent,
                attr=self.attr,
                ignore_excluded=self.ignore_excluded,
            )
            sent_start = getattr(sent, "start", 0)
            for size in range(1, self.windows):
                for i in range(0, len(offsets) - size):
                    begin_char, _, begin_i, _ = offsets[i]
                    _, end_char, _, end_i = offsets[i + size]
                    span_text = "##" + text[begin_char:end_char] + "##"
                    matches = self.ss_reader.retrieve(span_text)
                    for res in matches:
                        sim = _similarity(span_text, res, measure=self.measure)
                        for cui in self.syn2cuis[res]:
                            ents.append(
                                (cui, begin_i + sent_start, end_i + sent_start, sim)
                            )

        sorted_spans = sorted(ents, key=simstring_sort_key, reverse=True)
        results = []
        seen_tokens = set()
        for span in sorted_spans:
            # Check for end - 1 here because boundaries are inclusive
            span_tokens = set(range(span[1], span[2]))
            if not (span_tokens & seen_tokens):
                results.append(span)
                seen_tokens.update(span_tokens)
        results = sorted(results, key=lambda span: span[1])
        if as_spans:
            spans = [
                Span(root, span_data[1], span_data[2], span_data[0])
                for span_data in results
            ]
            return spans
        else:
            return [(self.vocab.strings[span[0]], span[1], span[2]) for span in results]


def _similarity(x: str, y: str, measure: SimilarityMeasure = SimilarityMeasure.dice):
    x_ngrams = {x[i : i + 3] for i in range(0, len(x) - 3)}
    y_ngrams = {y[i : i + 3] for i in range(0, len(y) - 3)}

    if measure == SimilarityMeasure.jaccard:
        return len(x_ngrams & y_ngrams) / (len(x_ngrams | y_ngrams))

    if measure == SimilarityMeasure.dice:
        return 2 * len(x_ngrams & y_ngrams) / (len(x_ngrams) + len(y_ngrams))

    if measure == SimilarityMeasure.cosine:
        return len(x_ngrams & y_ngrams) / sqrt(len(x_ngrams) * len(y_ngrams))

    if measure == SimilarityMeasure.overlap:
        return len(x_ngrams & y_ngrams)


def simstring_sort_key(span_data: Tuple[str, int, int, float]):
    return span_data[3], span_data[2] - span_data[1], -span_data[1]


@lru_cache(maxsize=128)
def get_text_and_offsets(
    doclike: Union[Span, Doc],
    attr: str = "TEXT",
    ignore_excluded: bool = True,
    ignore_space_tokens: bool = True,
) -> Tuple[str, List[Tuple[int, int, int]]]:
    """
    Align different representations of a `Doc` or `Span` object.

    Parameters
    ----------
    doclike : Doc
        spaCy `Doc` or `Span` object
    attr : str, optional
        Attribute to use, by default `"TEXT"`
    ignore_excluded : bool
        Whether to remove excluded tokens, by default True
    ignore_space_tokens : bool
        Whether to remove space tokens, by default False


    Returns
    -------
    Tuple[str, List[Tuple[int, int, int, int]]]
        The new clean text and offset tuples for each word giving the begin char indice
        of the word in the new text, the end char indice of its preceding word and the
        begin / end indices of the word in the original document
    """
    attr = attr.upper()
    attr = ATTRIBUTES.get(attr, attr)

    custom = attr.startswith("_")

    if custom:
        attr = attr[1:].lower()

    offsets = []

    cursor = 0

    text = []

    last = cursor
    last_i = 0
    for i, token in enumerate(doclike):
        if (not ignore_excluded or token.tag_ != "EXCLUDED") and (
            not ignore_space_tokens or token.tag_ != "SPACE"
        ):
            if custom:
                token_text = getattr(token._, attr)
            else:
                token_text = getattr(token, attr)

            # We add the cursor
            end = cursor + len(token_text)
            offsets.append((cursor, last, i, last_i + 1))

            cursor = end
            last = end
            last_i = i

            text.append(token_text)

            if token.whitespace_:
                cursor += 1
                text.append(" ")

    offsets.append((cursor, last, len(doclike), last_i + 1))

    return "".join(text), offsets



from edsnlp.core import registry

from . import patterns
from .accents import AccentsConverter

DEFAULT_CONFIG = dict(
    accents=patterns.accents,
)

create_component = registry.factory.register(
    "eds.accents",
    assigns=["token.norm"],
    deprecated=["accents"],
)(AccentsConverter)



from edsnlp.core import registry

from .cerebrovascular_accident import CerebrovascularAccidentMatcher
from .patterns import default_patterns

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    label="cerebrovascular_accident",
    span_setter={"ents": True, "cerebrovascular_accident": True},
)

create_component = registry.factory.register(
    "eds.cerebrovascular_accident",
    assigns=["doc.ents", "doc.spans"],
)(CerebrovascularAccidentMatcher)



from datetime import date, datetime

import pendulum
from spacy.tokens import Doc

if not Doc.has_extension("note_id"):
    Doc.set_extension("note_id", default=None)


def set_note_datetime(span, dt):
    if type(dt) is datetime:
        dt = pendulum.instance(dt)
    elif isinstance(dt, pendulum.DateTime):
        pass
    elif isinstance(dt, str):
        dt = pendulum.parse(dt)
    elif isinstance(dt, (int, float)):
        dt = pendulum.from_timestamp(dt)
    elif isinstance(dt, date):
        dt = pendulum.instance(datetime.fromordinal(dt.toordinal()))
    elif dt is None:
        pass
    else:
        raise ValueError(f"Cannot cast {dt} as a datetime")
    key = span._._get_key("note_datetime")
    span.doc.user_data[key] = dt


def get_note_datetime(doc):
    key = doc._._get_key("note_datetime")
    return doc.user_data.get(key, None)


if not Doc.has_extension("note_datetime"):
    Doc.set_extension(
        "note_datetime",
        getter=get_note_datetime,
        setter=set_note_datetime,
    )

if not Doc.has_extension("birth_datetime"):
    Doc.set_extension("birth_datetime", default=None)



import random
import re
from random import choice, randint
from random import random as randomf
from string import ascii_letters, ascii_lowercase

import pandas as pd
import pytest

from edsnlp.connectors.omop import OmopConnector

random.seed(10)


def random_word():
    n = randint(1, 20)
    return "".join(
        [choice(ascii_letters)] + [choice(ascii_lowercase) for _ in range(n)]
    )


def random_text():
    n = randint(30, 60)
    return " ".join([random_word() for _ in range(n)])


def random_note_nlp(text):

    ents = []

    for match in re.finditer(r"\w+", text):
        if randomf() > 0.8:
            ent = dict(
                start_char=match.start(),
                end_char=match.end(),
                lexical_variant=match.group(),
                note_nlp_source_value=random_word().lower(),
                negation=randomf() > 0.5,
            )
            ents.append(ent)

    return ents


@pytest.fixture
def note():

    df = pd.DataFrame(dict(note_text=[random_text() for _ in range(10)]))
    df["note_id"] = range(len(df))
    df["note_datetime"] = "2021-10-19"

    return df


@pytest.fixture
def note_nlp(note):

    df = note.copy()
    df["ents"] = df.note_text.apply(random_note_nlp)
    df = df.explode("ents")
    df = pd.concat([df, df.ents.apply(pd.Series)], axis=1)
    df = df.drop(columns=["note_text", "ents", "note_datetime"])
    df["note_nlp_id"] = range(len(df))

    return df


@pytest.fixture
def omop(blank_nlp) -> OmopConnector:
    blank_nlp.add_pipe("negation")
    return OmopConnector(blank_nlp)


@pytest.fixture
def docs(omop: OmopConnector, note, note_nlp):
    return omop.omop2docs(note, note_nlp, extensions=["negation"])


def test_omop2docs(docs, note, note_nlp):

    lexical_variants = note_nlp.groupby("note_id")["lexical_variant"].agg(list)

    for doc, text, lvs in zip(docs, note.note_text, lexical_variants):
        assert doc.text == text
        assert len(doc.ents) == len(lvs)

        for ent, lv in zip(doc.ents, lvs):
            assert ent.text == lv


def test_docs2omop(omop: OmopConnector, docs):
    note, note_nlp = omop.docs2omop(docs, extensions=["negation"])

    lexical_variants = note_nlp.groupby("note_id")["lexical_variant"].agg(list)

    for doc, text, lvs in zip(docs, note.note_text, lexical_variants):
        assert doc.text == text
        assert len(doc.ents) == len(lvs)

        for ent, lv in zip(doc.ents, lvs):
            assert ent.text == lv


def test_roundtrip(omop: OmopConnector, docs, note, note_nlp):
    note2, note_nlp2 = omop.docs2omop(docs, extensions=["negation"])

    assert (note2 == note[note2.columns]).all().all()
    assert (note_nlp2 == note_nlp[note_nlp2.columns]).all().all()



from __future__ import annotations

import json
import os
import warnings
from collections import defaultdict
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
)

import foldedtensor as ft
import torch
import torch.nn.functional as F
from spacy.tokens import Doc, Span
from typing_extensions import Literal, NotRequired, TypedDict

import edsnlp.data
from edsnlp.core import PipelineProtocol
from edsnlp.core.torch_component import BatchInput, BatchOutput, TorchComponent
from edsnlp.pipes.base import BaseComponent
from edsnlp.pipes.trainable.embeddings.typing import (
    SpanEmbeddingComponent,
)
from edsnlp.pipes.trainable.layers.metric import Metric
from edsnlp.utils.collections import (
    batch_compress_dict,
    decompress_dict,
    ld_to_dl,
)
from edsnlp.utils.span_getters import get_spans, get_spans_with_group

SpanLinkerBatchInput = TypedDict(
    "SpanLinkerBatchInput",
    {
        "embedding": BatchInput,
        "span_labels": torch.Tensor,
        "concepts": NotRequired[torch.Tensor],
    },
)
"""
embeds: torch.FloatTensor
    Token embeddings to predict the tags from
mask: torch.BoolTensor
    Mask of the sequences
spans: torch.LongTensor
    2d tensor of n_spans * (doc_idx, ner_label_idx, begin, end)
targets: NotRequired[List[torch.LongTensor]]
    list of 2d tensor of n_spans * n_combinations (1 hot)
"""


class TrainableSpanLinker(
    TorchComponent[BatchOutput, SpanLinkerBatchInput], BaseComponent
):
    """
    The `eds.span_linker` component is a trainable span concept predictor, typically
    used to match spans in the text with concepts in a knowledge base. This task is
    known as "Entity Linking", "Named Entity Disambiguation" or "Normalization" (the
    latter is mostly used in the biomedical machine learning community).

    !!! warning "Entity Linking vs Named Entity Recognition"

        Entity Linking is the task of linking existing entities to their concept in a
        knowledge base, while Named Entity Recognition is the task of detecting spans in
        the text that correspond to entities. The `eds.span_linker` component should
        therefore be used after the Named Entity Recognition step (e.g. using the
        `eds.ner_crf` component).

    How it works
    ------------

    To perform this task, this components compare the embedding of a given query span
    (e.g. "aspirin") with the embeddings in the knowledge base, where
    each embedding represents a concept (e.g. "B01AC06"), and selects the most similar
    embedding and returns its concept id. This comparison is done using either:

    - the cosine similarity between the input and output embeddings (recommended)
    - a simple dot product

    We filter out the concepts that are not relevant for a given query by using groups.
    For each span to link, we use its label to select a group of concepts to compare
    with. For example, if the span is labeled as "drug", we only compare it with
    concepts that are drugs. These concepts groups are inferred from the training data
    when running the `post_init` method, or can be provided manually using the
    `pipe.update_concepts(concepts, mapping, [embeddings])` method. If a label is not
    found in the mapping, the span is compared with all concepts.

    We support comparing entity queries against two kind of references : either the
    embeddings of the concepts themselves (`reference_mode = "concept"`), or the
    embeddings of the synonyms of the concepts (`reference_mode = "synonym"`).

    ### Synonym similarity

    When performing span linking in `synonym` mode, the span linker embedding matrix
    contains one embedding vector per concept per synonym, and each embedding maps to
    the concept of its synonym. This mode is slower and more memory intensive, since you
    have to store multiple embeddings per concept, but it can yield good results
    in zero-shot scenarios (see example below).

    <figure markdown>
      ![Synonym similarity](/assets/images/synonym_span_linker.png)
      <figcaption>Entity linking based on synonym similarity</figcaption>
    </figure>

    ### Concept similarity

    In `concept` mode, the span linker embedding matrix contains one embedding vector
    per concept : imagine a single vector that approximately averages all the synonyms
    of a concept (e.g. B01AC06 = average of "aspirin", "acetyl-salicylic acid", etc.).
    This mode is faster and more memory efficient, but usually requires that the
    concept weights are fine-tuned.

    <figure markdown>
      ![Concept similarity](/assets/images/class_span_linker.png)
      <figcaption>Entity linking based on concept similarity</figcaption>
    </figure>

    Examples
    --------

    Here is how you can use the `eds.span_linker` component to link spans without
    training, in `synonym` mode. You will still need to pre-compute the embeddings of
    the target synonyms.


    First, initialize the component:

    ```{ .python .no-check }
    import pandas as pd
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.span_linker(
            rescale=20.0,
            threshold=0.8,
            metric="cosine",
            reference_mode="synonym",
            probability_mode="sigmoid",
            embedding=eds.span_pooler(
                span_getter=["ents"],
                hidden_size=128,
                embedding=eds.transformer(
                    span_getter=["ents"],
                    model="prajjwal1/bert-tiny",
                    window=128,
                    stride=96,
                ),
            ),
        ),
        name="linker",
    )
    ```

    We will assume you have a list of synonyms with their concept and label with the
    columns:

    - `STR`: synonym text
    - `CUI`: concept id
    - `GRP`: label.

    All we need to do is to initialize the component with the synonyms and that's it !
    Since we have set `init_weights` to True, and we are in `synonym` mode, the
    embeddings of the synonyms will be stored in the component and used to compute the
    similarity scores

    ```{ .python .no-check }
    synonyms_df = pd.read_csv("synonyms.csv")

    def make_doc(row):
        doc = nlp.make_doc(row["STR"])
        span = doc[:]
        span.label_ = row["GRP"]
        doc.ents = [span]
        span._.cui = row["CUI"]
        return doc

    nlp.post_init(
        edsnlp.data.from_pandas(
            synonyms_df,
            converter=make_doc,
        )
    )
    ```

    Now, you can now use it in a text:
    ```{ .python .no-check }
    doc = nlp.make_doc("Aspirin is a drug")
    span = doc[0:1]  # "Aspirin"
    span.label_ = "Drug"
    doc.ents = [span]

    doc = nlp(doc)
    print(doc.ents[0]._.cui)  # "B01AC06"
    ```

    To use the `eds.span_linker` component in `class` mode, we refer to the following
    repository: [deep_multilingual_normalization](
    https://github.com/percevalw/deep_multilingual_normalization) based on the work
    of [@wajsburt2021medical].

    Parameters
    ----------
    nlp: PipelineProtocol
        Spacy vocabulary
    name: str
        Name of the component
    embedding : SpanEmbeddingComponent
        The word embedding component
    metric : Literal["cosine", "dot"] = "cosine"
        Whether to compute the cosine similarity between the input and output embeddings
        or the dot product.
    rescale : float
        Rescale the output cosine similarities by a constant factor.
    threshold : float
        Threshold probability to consider a concept as valid
    attribute : str
        The attribute to store the concept id
    reference_mode : Literal["concept", "synonym"]
        Whether to compare the embeddings with the concepts embeddings (one per concept)
        or the synonyms embeddings (one per concept per synonym). See above for more
        details.
    probability_mode : Literal["softmax", "sigmoid"]
        Whether to compute the probabilities using a softmax or a sigmoid function.
        This will also determine the loss function to use, either cross-entropy or
        binary cross-entropy.

        !!! warning "Subsetting the concepts"

            The probabilities returned in `softmax` mode depend on the number of
            concepts (as an extreme cas, if you have only one concept, its softmax
            probability will always be 1). This is why we recommend using the `sigmoid`
            mode in which the probabilities are computed independently for each concept.
    init_weights : bool
        Whether to initialize the weights of the component with the embeddings of
        the entities of the docs provided to the `post_init` method.
        How this is done depends on the `reference_mode` parameter:

        - `concept`: the embeddings are averaged
        - `synonym`: the embeddings are stored as is

        By default, this is set to `True` if `reference_mode` is `synonym`, and
        `False` otherwise.

    Authors and citation
    --------------------
    The `eds.span_linker` component was developed by AP-HP's Data Science team.

    The deep learning concept-based architecture was adapted from
    [@wajsburt2021medical].
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        name: str = "span_linker",
        *,
        embedding: SpanEmbeddingComponent,
        metric: Literal["cosine", "dot"] = "cosine",
        rescale: float = 20,
        threshold: float = 0.5,
        attribute: str = "cui",
        reference_mode: Literal["concept", "synonym"] = "concept",
        probability_mode: Literal["softmax", "sigmoid"] = "sigmoid",
        init_weights: bool = True,
    ):
        self.attribute = attribute

        super().__init__(nlp, name)

        self.embedding = embedding
        self.concepts: List = []
        self.qualifiers = ["_." + self.attribute]
        self.concepts_to_idx: Dict[str, int] = {}
        self.span_labels_to_idx: Dict[str, int] = {}
        self.threshold = threshold
        self.reference_mode = reference_mode
        self.probability_mode = probability_mode
        self.init_weights = init_weights
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            self.classifier = Metric(
                in_features=self.embedding.output_size,
                out_features=0,
                num_groups=0,
                rescale=rescale,
                metric=metric,
            )

    @property
    def span_getter(self):
        return self.embedding.span_getter

    def to_disk(self, path, *, exclude=set()):
        repr_id = object.__repr__(self)
        if repr_id in exclude:
            return
        # This will receive the directory path + /my_component
        # We save the bindings as a pickle file since values can be arbitrary objects
        os.makedirs(path, exist_ok=True)
        data_path = path / "vocab.json"
        with open(data_path, "w") as file:
            json.dump(
                {
                    "concepts": self.concepts,
                    "span_labels": list(self.span_labels_to_idx.keys()),
                },
                file,
            )
        return super().to_disk(path, exclude=exclude)

    def from_disk(self, path, exclude=tuple()):
        repr_id = object.__repr__(self)
        if repr_id in exclude:
            return
        # This will receive the directory path + /my_component
        data_path = path / "vocab.json"
        with open(data_path, "r") as f:
            data = json.load(f)
        self.span_labels_to_idx = {lab: i for i, lab in enumerate(data["span_labels"])}
        self.classifier.groups.data = torch.zeros(
            len(self.span_labels_to_idx),
            len(data["concepts"]),
            dtype=torch.bool,
        )
        self.update_concepts(concepts=data["concepts"])
        self.set_extensions()
        super().from_disk(path, exclude=exclude)

    def set_extensions(self):
        if not Span.has_extension(self.attribute):
            Span.set_extension(self.attribute, default=None)
        if not Span.has_extension("prob"):
            # Builtin mutable defaults will be copied the first time they're accessed
            Span.set_extension("prob", default={})
        super().set_extensions()

    def update_concepts(
        self,
        *,
        concepts: Optional[Sequence[str]] = None,
        mapping: Optional[Dict[str, Iterable[str]]] = None,
        embeds: Optional[torch.Tensor] = None,
    ):
        old_concepts_to_idx = self.concepts_to_idx
        device = self.device
        if mapping is not None:
            mapping = {k: set(v) for k, v in mapping.items()}
            concepts = (
                sorted(set(cui for cuis in mapping.values() for cui in cuis))
                if concepts is None
                else concepts
            )

        assert concepts is not None, "You must provide either `concepts` or `mapping`"

        self.concepts_to_idx = {
            cui: idx for idx, cui in enumerate(sorted(set(concepts)))
        }

        new_lin = torch.nn.Linear(
            self.classifier.in_features,
            len(concepts)
            if self.reference_mode == "synonym"
            else len(self.concepts_to_idx),
        ).to(device)
        if embeds is not None:
            if self.reference_mode == "synonym":
                embeds = embeds.to(new_lin.weight)
                new_lin.weight.data = embeds
            else:
                old_weight_norm = (
                    self.classifier.weight.norm(dim=-1).mean().item()
                    if self.classifier.weight.shape[0]
                    else new_lin.weight.norm(dim=-1).mean().item()
                )
                cui_indices = torch.as_tensor(
                    [self.concepts_to_idx[c] for c in concepts],
                    dtype=torch.long,
                ).to(device)
                new_lin.weight.data.zero_().index_add_(
                    0, cui_indices, embeds.to(new_lin.weight)
                )
                del embeds
                new_lin.weight.data = (
                    F.normalize(
                        new_lin.weight.data
                        / cui_indices.bincount(minlength=len(self.concepts_to_idx))
                        .to(device)
                        .unsqueeze(-1)
                    )
                    * old_weight_norm
                )
        else:
            common = sorted(set(self.concepts_to_idx) & set(old_concepts_to_idx))
            old_index = [old_concepts_to_idx[label] for label in common]
            new_index = [self.concepts_to_idx[label] for label in common]
            new_lin.weight.data[new_index] = self.classifier.weight.data[old_index]

        self.concepts = (
            concepts if self.reference_mode == "synonym" else list(self.concepts_to_idx)
        )
        self.classifier.weight.data = new_lin.weight.data
        self.classifier.out_features = self.classifier.weight.shape[0]

        if mapping is not None:
            mapping = {k: set(v) for k, v in mapping.items()}
            self.classifier.groups.data = torch.tensor(
                [[True] * len(self.concepts)]
                + [
                    [concept in label_concepts for concept in self.concepts]
                    for span_label, label_concepts in mapping.items()
                ],
                device=device,
            )
            span_labels = [None] + list(mapping.keys())
            span_labels_to_idx = {label: idx for idx, label in enumerate(span_labels)}
            self.span_labels_to_idx = span_labels_to_idx

    def compute_init_features(
        self,
        docs: Iterable[Doc],
        with_embeddings=True,
    ) -> Tuple[List[str], List[str], Optional[torch.Tensor]]:
        embedding = self.embedding

        def prepare_batch(docs, device=None):
            res = {}
            if self.init_weights:
                device = device or embedding.device
                batch = [embedding.preprocess(doc) for doc in docs]
                spans = [span for prep in batch for span in prep["$spans"]]
                batch = decompress_dict(list(batch_compress_dict(batch)))
                inputs = embedding.collate(batch)
                inputs = embedding.batch_to_device(inputs, device=device)
                res["inputs"] = inputs
            else:
                spans = [
                    span for doc in docs for span in get_spans(doc, self.span_getter)
                ]
            batch_concepts = [s._.get(self.attribute) for s in spans]
            batch_labels = [s.label_ for s in spans]
            res.update(
                {
                    "concepts": batch_concepts,
                    "labels": batch_labels,
                }
            )
            return res

        def run_forward(batch):
            res = {
                "concepts": batch["concepts"],
                "labels": batch["labels"],
            }
            if self.init_weights:
                res["embeds"] = F.normalize(
                    embedding.module_forward(batch["inputs"])["embeddings"]
                ).cpu()
            return res

        # Loop over the data to compute the embeddings and collect the concepts
        docs = edsnlp.data.from_iterable(docs)
        results = ld_to_dl(
            docs.map_gpu(
                forward=run_forward,
                prepare_batch=prepare_batch,
            )
            if self.init_weights
            else docs.map_batches(prepare_batch)
        )
        all_embeds = (
            torch.cat(results["embeds"], dim=0).to(self.device)
            if "embeds" in results
            else None
        )
        all_concepts = [c for batch in results["concepts"] for c in batch]
        all_labels = [c for batch in results["labels"] for c in batch]
        if not len(all_concepts):
            warnings.warn("Did not find any concept when scanning the gold data.")

        return all_concepts, all_labels, all_embeds

    def post_init(self, gold_data: Iterable[Doc], exclude: Set[str]):
        super().post_init(gold_data, exclude=exclude)

        concepts, labels, embeds = self.compute_init_features(
            gold_data, with_embeddings=self.init_weights
        )
        labels_to_cuis = defaultdict(set)
        for label, cui in zip(labels, concepts):
            if cui is not None:
                labels_to_cuis[label].add(cui)

        self.update_concepts(
            concepts=concepts,
            mapping=labels_to_cuis,
            embeds=embeds,
        )

    def preprocess(self, doc: Doc) -> Dict[str, Any]:
        embedding = self.embedding.preprocess(doc)

        # "$" prefix means this field won't be accessible from the outside.
        spans = embedding["$spans"]
        return {
            "embedding": embedding,
            "span_labels": [
                self.span_labels_to_idx.get(span.label_, 0) for span in spans
            ],
        }

    def preprocess_supervised(self, doc: Doc) -> Dict[str, Any]:
        preps = self.preprocess(doc)
        spans = preps["embedding"]["$spans"]
        return {
            **preps,
            "concepts": [self.concepts_to_idx.get(span._.cui, -100) for span in spans],
        }

    def collate(self, batch: Dict[str, Sequence[Any]]) -> SpanLinkerBatchInput:
        span_labels = ft.as_folded_tensor(
            batch["span_labels"],
            dtype=torch.long,
            full_names=("sample", "span"),
            data_dims=("span",),
        ).as_tensor()
        collated: SpanLinkerBatchInput = {
            "span_labels": span_labels,
            "embedding": self.embedding.collate(batch["embedding"]),
        }
        if "concepts" in batch:
            collated["concepts"] = ft.as_folded_tensor(
                batch["concepts"],
                dtype=torch.long,
                full_names=("sample", "span"),
                data_dims=("span",),
            ).as_tensor()
        return collated

    def forward(self, inputs: SpanLinkerBatchInput) -> BatchOutput:
        span_embeds = self.embedding.module_forward(inputs["embedding"])["embeddings"]

        span_labels = inputs["span_labels"]
        targets = inputs.get("concepts")

        scores = self.classifier(
            inputs=span_embeds,
            group_indices=span_labels,
        )
        concepts = loss = top_probs = None
        if targets is not None:
            #  TRAINING
            num_classes = self.classifier.weight.shape[0]
            loss = (
                F.binary_cross_entropy_with_logits(
                    scores,
                    F.one_hot(targets, num_classes=num_classes).float(),
                    reduction="sum",
                )
                / num_classes
                if self.probability_mode == "sigmoid"
                else F.cross_entropy(scores, targets, reduction="sum")
            )
        else:
            # PREDICTION
            probs = (
                torch.sigmoid(scores)
                if self.probability_mode == "sigmoid"
                else torch.softmax(scores, -1)
            )
            top_probs, concepts = probs.max(-1)

        return {
            "loss": loss,
            "concepts": concepts,
            "probs": top_probs,
        }

    def postprocess(self, docs: Sequence[Doc], res: BatchOutput) -> Sequence[Doc]:
        spans = [
            (s, group)
            for doc in docs
            for s, group in get_spans_with_group(doc, self.span_getter)
        ]
        ents = {doc: {} for doc in docs}
        for (span, group), concept_idx, prob in zip(
            spans,
            res["concepts"].tolist(),
            res["probs"].tolist(),
        ):
            span._.set(
                self.attribute,
                self.concepts[concept_idx] if prob >= self.threshold else None,
            )
            span._.prob[self.attribute] = {self.concepts[concept_idx]: prob}
            if group == "ents":
                ents[span.doc][span.start] = span
        for doc, doc_ents in ents.items():
            if doc_ents:
                doc.ents = {
                    **{span.start: span for span in doc.ents},
                    **doc_ents,
                }.values()
        return docs



from edsnlp.core import registry

from .hypothesis import HypothesisQualifier

DEFAULT_CONFIG = dict(
    pseudo=None,
    preceding=None,
    following=None,
    verbs_eds=None,
    verbs_hyp=None,
    termination=None,
    attr="NORM",
    span_getter=None,
    on_ents_only=True,
    within_ents=False,
    explain=False,
)

create_component = registry.factory.register(
    "eds.hypothesis",
    assigns=["span._.hypothesis"],
    deprecated=["hypothesis"],
)(HypothesisQualifier)



main_pattern = dict(
    source="main",
    regex=[
        r"leucemie",
        r"(syndrome.)?myeloproliferatif",
        r"m[yi]eloprolifer",
    ],
    exclude=dict(
        regex=[
            "plasmocyte",
            "benin",
            "benign",
        ],
        window=5,
    ),
    regex_attr="NORM",
)

acronym = dict(
    source="acronym",
    regex=[
        r"\bLAM\b",
        r"\bLAM.?[0-9]",
        r"\bLAL\b",
        r"\bLMC\b",
        r"\bLCE\b",
        r"\bLMM[JC]\b",
        r"\bLCN\b",
        r"\bAREB\b",
        r"\bAPMF\b",
        r"\bLLC\b",
        r"\bSMD\b",
        r"LA my[éèe]lomonocytaire",
    ],
    regex_attr="TEXT",
    exclude=dict(
        regex="anti",
        window=-20,
    ),
)

other = dict(
    source="other",
    regex=[
        r"myelofibrose",
        r"vaquez",
        r"thrombocytemie.{1,3}essentielle",
        r"splenomegalie.{1,3}myeloide",
        r"mastocytose.{1,5}maligne",
        r"polyglobulie.{1,10}essentielle",
        r"letterer.?siwe",
        r"anemie.refractaire.{1,20}blaste",
        r"m[iy]elod[iy]splasi",
        r"syndrome.myelo.?dysplasique",
    ],
    regex_attr="NORM",
)

default_patterns = [
    main_pattern,
    acronym,
    other,
]



import functools
from typing import Any, List


def rgetattr(obj: Any, attr: str, *args: List[Any]) -> Any:
    """
    Get attribute recursively

    Parameters
    ----------
    obj : Any
        An object
    attr : str
        The name of the attribute to get. Can contain dots.
    """

    def _getattr(obj, attr):
        return None if obj is None else getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))



import re
from typing import List, Optional, Tuple, Union

import pydantic
import regex
from pydantic import BaseModel, Extra, validator

from edsnlp.matchers.utils import ListOrStr

Flags = Union[re.RegexFlag, int]
Window = Union[
    Tuple[int, int],
    List[int],
    int,
]


def normalize_window(cls, v):
    if isinstance(v, list):
        assert (
            len(v) == 2
        ), "`window` should be a tuple/list of two integer, or a single integer"
        v = tuple(v)
    if isinstance(v, int):
        assert v != 0, "The provided `window` should not be 0"
        if v < 0:
            return (v, 0)
        if v > 0:
            return (0, v)
    assert v[0] < v[1], "The provided `window` should contain at least 1 token"
    return v


class AssignDict(dict):
    """
    Custom dictionary that overrides the __setitem__ method
    depending on the reduce_mode
    """

    def __init__(self, reduce_mode: dict):
        super().__init__()
        self.reduce_mode = reduce_mode
        self._setitem_ = self.__setitem_options__()

    def __missing__(self, key):
        return (
            {
                "span": [],
                "value_span": [],
                "value_text": [],
            }
            if self.reduce_mode[key] is None
            else {}
        )

    def __setitem__(self, key, value):
        self._setitem_[self.reduce_mode[key]](key, value)

    def __setitem_options__(self):
        def keep_list(key, value):
            old_values = self.__getitem__(key)
            value["span"] = old_values["span"] + [value["span"]]
            value["value_span"] = old_values["value_span"] + [value["value_span"]]
            value["value_text"] = old_values["value_text"] + [value["value_text"]]

            dict.__setitem__(self, key, value)

        def keep_first(key, value):
            old_values = self.__getitem__(key)
            if (
                old_values.get("span") is None
                or value["span"].start <= old_values["span"].start
            ):
                dict.__setitem__(self, key, value)

        def keep_last(key, value):
            old_values = self.__getitem__(key)
            if (
                old_values.get("span") is None
                or value["span"].start >= old_values["span"].start
            ):
                dict.__setitem__(self, key, value)

        return {
            None: keep_list,
            "keep_first": keep_first,
            "keep_last": keep_last,
        }


class SingleExcludeModel(BaseModel):
    regex: ListOrStr = []
    window: Window
    limit_to_sentence: Optional[bool] = True
    regex_flags: Optional[Flags] = None
    regex_attr: Optional[str] = None

    @validator("regex")
    def exclude_regex_validation(cls, v):
        if isinstance(v, str):
            v = [v]
        return v

    _normalize_window = validator("window", allow_reuse=True)(normalize_window)


class ExcludeModel:
    @classmethod
    def item_to_list(cls, v, config):
        if not isinstance(v, list):
            v = [v]
        return [pydantic.parse_obj_as(SingleExcludeModel, x) for x in v]

    @classmethod
    def __get_validators__(cls):
        yield cls.item_to_list


class SingleAssignModel(BaseModel):
    name: str
    regex: str
    window: Window
    limit_to_sentence: Optional[bool] = True
    regex_flags: Optional[Flags] = None
    regex_attr: Optional[str] = None
    replace_entity: bool = False
    reduce_mode: Optional[str] = None

    @validator("regex")
    def check_single_regex_group(cls, pat):
        compiled_pat = regex.compile(
            pat
        )  # Using regex to allow multiple fgroups with same name
        n_groups = compiled_pat.groups
        assert n_groups == 1, (
            "The pattern {pat} should have only one capturing group, not {n_groups}"
        ).format(
            pat=pat,
            n_groups=n_groups,
        )

        return pat

    _normalize_window = validator("window", allow_reuse=True)(normalize_window)


class AssignModel:
    @classmethod
    def item_to_list(cls, v, config):
        if not isinstance(v, list):
            v = [v]
        return [pydantic.parse_obj_as(SingleAssignModel, x) for x in v]

    @classmethod
    def name_uniqueness(cls, v, config):
        names = [item.name for item in v]
        assert len(names) == len(set(names)), "Each `name` field should be unique"
        return v

    @classmethod
    def replace_uniqueness(cls, v, config):
        replace = [item for item in v if item.replace_entity]
        assert (
            len(replace) <= 1
        ), "Only 1 assign element can be set with `replace_entity=True`"
        return v

    @classmethod
    def __get_validators__(cls):
        yield cls.item_to_list
        yield cls.name_uniqueness
        yield cls.replace_uniqueness


class SingleConfig(BaseModel, extra=Extra.forbid):
    source: str
    terms: ListOrStr = []
    regex: ListOrStr = []
    regex_attr: Optional[str] = None
    regex_flags: Union[re.RegexFlag, int] = None
    exclude: Optional[ExcludeModel] = []
    assign: Optional[AssignModel] = []


class FullConfig:
    @classmethod
    def pattern_to_list(cls, v, config):
        if not isinstance(v, list):
            v = [v]
        return [pydantic.parse_obj_as(SingleConfig, item) for item in v]

    @classmethod
    def source_uniqueness(cls, v, config):
        sources = [item.source for item in v]
        assert len(sources) == len(set(sources)), "Each `source` field should be unique"
        return v

    @classmethod
    def __get_validators__(cls):
        yield cls.pattern_to_list
        yield cls.source_uniqueness



from pytest import mark

from edsnlp.matchers.simstring import SimstringMatcher


def test_simstring_matcher(doc, nlp):
    matcher = SimstringMatcher(nlp.vocab, attr="TEXT")

    matcher.build_patterns(
        nlp,
        {
            "patient": ["patient"],
            "locomotion": ["locomotions"],
        },
    )

    matches = [m.text for m in matcher(doc, as_spans=True)]

    assert matches == ["patient", "locomotion", "patient"]


def test_with_normalizer(blank_nlp):
    blank_nlp.add_pipe("eds.normalizer")
    pattern = "matching"
    matcher = SimstringMatcher(
        blank_nlp.vocab,
        attr="NORM",
        threshold=0.75,
        measure="dice",
        ignore_space_tokens=True,
        ignore_excluded=True,
    )

    matcher.build_patterns(
        blank_nlp,
        {
            "test": [pattern],
            "C220": ["carcinome hépatocellulaire", "carc. hépatocellulaire"],
            "N02BE01": ["paracetamol"],
        },
    )

    texts = (
        ("Ceci est un test de matching", ["matching"]),
        ("Ceci est un test de matchings", ["matchings"]),
        ("On prescrit du paracétomol      , un medicament.", ["paracétomol"]),
        (
            "Le patient a un carcinome\nhépatacellulaire !",
            ["carcinome\nhépatacellulaire"],
        ),
    )

    for text, ents in texts:
        doc = blank_nlp(text)
        matches = list(matcher(doc))
        assert len(matches) > 0

        assert sorted([m.text for m in matcher(doc[2:], as_spans=True)]) == ents
        assert sorted([doc[s:e].text for _, s, e in matcher(doc[2:])]) == ents

        assert sorted([m.text for m in matcher(doc, as_spans=True)]) == ents
        assert sorted([doc[s:e].text for _, s, e in matcher(doc)]) == ents


@mark.parametrize("measure", ["dice", "cosine", "jaccard", "overlap"])
def test_without_normalizer(blank_nlp, measure):
    pattern = "matching"
    matcher = SimstringMatcher(
        blank_nlp.vocab, attr="NORM", threshold=0.6, measure=measure
    )

    matcher.build_patterns(
        blank_nlp,
        {
            "test": [pattern],
            "C220": ["carcinome hépatocellulaire", "carc. hépatocellulaire"],
            "N02BE01": ["paracétamol"],
        },
    )

    texts = (
        ("Ceci est un test de matching", ["matching"]),
        ("Ceci est un test de matchings", ["matchings"]),
        ("On prescrit du paracétomol, un médicament.", ["paracétomol"]),
        (
            "Le patient a un carcinome hépatacellulaire !",
            ["carcinome hépatacellulaire"],
        ),
    )

    for text, ents in texts:
        doc = blank_nlp(text)
        matches = list(matcher(doc))
        assert len(matches) > 0

        assert sorted([m.text for m in matcher(doc[2:], as_spans=True)]) == ents
        assert sorted([doc[s:e].text for _, s, e in matcher(doc[2:])]) == ents

        assert sorted([m.text for m in matcher(doc, as_spans=True)]) == ents
        assert sorted([doc[s:e].text for _, s, e in matcher(doc)]) == ents



from itertools import chain
from typing import Any, Dict, List, Optional

from spacy.tokens import Doc, Span
from typing_extensions import Literal

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.phrase import EDSPhraseMatcher
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.matchers.simstring import SimstringMatcher
from edsnlp.matchers.utils import Patterns
from edsnlp.pipes.base import BaseNERComponent, SpanSetterArg


class TerminologyMatcher(BaseNERComponent):
    r"""
    EDS-NLP simplifies the terminology matching process by exposing a `eds.terminology`
    pipeline that can match on terms or regular expressions.

    The terminology matcher is very similar to the
    [generic matcher][edsnlp.pipes.core.matcher.factory.create_component],
    although the use case differs slightly. The generic matcher is designed to extract
    any entity, while the terminology matcher is specifically tailored towards high
    volume terminologies.

    There are some key differences:

    1. It labels every matched entity to the same value, provided to the pipeline
    2. The keys provided in the `regex` and `terms` dictionaries are used as the
       `kb_id_` of the entity, which handles fine-grained labelling

    For instance, a terminology matcher could detect every drug mention under the
    top-level label `drug`, and link each individual mention to a given drug through
    its `kb_id_` attribute.

    Examples
    --------
    Let us redefine the pipeline :

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")

    terms = dict(
        covid=["coronavirus", "covid19"],  # (1)
        flu=["grippe saisonnière"],  # (2)
    )

    regex = dict(
        covid=r"coronavirus|covid[-\s]?19|sars[-\s]cov[-\s]2",  # (3)
    )

    nlp.add_pipe(
        eds.terminology(
            label="disease",
            terms=terms,
            regex=regex,
            attr="LOWER",
        ),
    )
    ```

    1. Every key in the `terms` dictionary is mapped to a concept.
    2. The `eds.matcher` pipeline expects a list of expressions, or a single expression.
    3. We can also define regular expression patterns.

    This snippet is complete, and should run as is.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    terms : Optional[Patterns]
        A dictionary of terms.
    regex : Optional[Patterns]
        A dictionary of regular expressions.
    attr : str
        The default attribute to use for matching.
        Can be overridden using the `terms` and `regex` configurations.
    ignore_excluded : bool
        Whether to skip excluded tokens (requires an upstream
        pipeline to mark excluded tokens).
    ignore_space_tokens: bool
        Whether to skip space tokens during matching.
    term_matcher: Literal["exact", "simstring"]
        The matcher to use for matching phrases ?
        One of (exact, simstring)
    term_matcher_config: Dict[str,Any]
        Parameters of the matcher class
    label: str
        Label name to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Patterns, be they `terms` or `regex`, are defined as dictionaries where keys become
    the `kb_id_` of the extracted entities. Dictionary values are either a single
    expression or a list of expressions that match the concept (see [example](#usage)).

    Authors and citation
    --------------------
    The `eds.terminology` pipeline was developed by AP-HP's Data Science team.
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "terminology",
        *,
        terms: Optional[Patterns] = None,
        regex: Optional[Patterns] = None,
        attr: str = "TEXT",
        ignore_excluded: bool = False,
        ignore_space_tokens: bool = False,
        term_matcher: Literal["exact", "simstring"] = "exact",
        term_matcher_config: Optional[Dict[str, Any]] = None,
        label,
        span_setter: SpanSetterArg = {"ents": True},
    ):
        self.label = label

        super().__init__(nlp=nlp, name=name, span_setter=span_setter)

        if terms is None and regex is None:
            raise ValueError(
                "You must provide either `terms` or `regex` to the matcher."
            )

        terms = terms or {}
        regex = regex or {}

        self.attr = attr

        if term_matcher == "exact":
            self.phrase_matcher = EDSPhraseMatcher(
                vocab=nlp.vocab,
                attr=attr,
                ignore_excluded=ignore_excluded,
                ignore_space_tokens=ignore_space_tokens,
                **(term_matcher_config or {}),
            )
        elif term_matcher == "simstring":
            self.phrase_matcher = SimstringMatcher(
                vocab=nlp.vocab,
                attr=attr,
                ignore_excluded=ignore_excluded,
                ignore_space_tokens=ignore_space_tokens,
                **(term_matcher_config or {}),
            )
        else:
            raise ValueError(
                f"Algorithm {repr(term_matcher)} does not belong to"
                f" known matchers [exact, simstring]."
            )

        self.regex_matcher = RegexMatcher(
            attr=attr,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
        )

        self.phrase_matcher.build_patterns(nlp=nlp, terms=terms, progress=True)
        self.regex_matcher.build_patterns(regex=regex)

        self.set_extensions()

    def set_extensions(self) -> None:
        super().set_extensions()
        if not Span.has_extension(self.label):
            Span.set_extension(self.label, default=None)

    def process(self, doc: Doc) -> List[Span]:
        """
        Find matching spans in doc.

        Post-process matches to account for terminology.

        Parameters
        ----------
        doc:
            spaCy Doc object.

        Returns
        -------
        spans:
            List of Spans returned by the matchers.
        """

        matches = self.phrase_matcher(doc, as_spans=True)
        regex_matches = self.regex_matcher(doc, as_spans=True)

        for match in chain(matches, regex_matches):
            span = Span(
                doc=match.doc,
                start=match.start,
                end=match.end,
                label=self.label,
                kb_id=match.label,
            )
            span._.set(self.label, match.label_)
            yield span

    def __call__(self, doc: Doc) -> Doc:
        """
        Adds spans to document.

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        doc:
            spaCy Doc object, annotated for extracted terms.
        """
        matches = self.process(doc)

        self.set_spans(doc, matches)

        return doc



from typing import Optional, Sequence

import torch
from typing_extensions import Literal, TypedDict

from edsnlp.core.pipeline import Pipeline
from edsnlp.core.torch_component import BatchInput
from edsnlp.pipes.trainable.embeddings.typing import (
    WordEmbeddingBatchOutput,
    WordEmbeddingComponent,
)
from edsnlp.pipes.trainable.layers.text_cnn import TextCnn
from edsnlp.utils.torch import ActivationFunction

TextCnnBatchInput = TypedDict(
    "TextCnnBatchInput",
    {
        "embeddings": torch.Tensor,
        "mask": torch.Tensor,
    },
)


class TextCnnEncoder(WordEmbeddingComponent):
    """
    The `eds.text_cnn` component is a simple 1D convolutional network to contextualize
    word embeddings (as computed by the `embedding` component passed as argument).

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : str
        The name of the component
    embedding : TorchComponent[WordEmbeddingBatchOutput, BatchInput]
        Embedding module to apply to the input
    output_size : Optional[int]
        Size of the output embeddings
        Defaults to the `input_size`
    out_channels : int
        Number of channels
    kernel_sizes : Sequence[int]
        Window size of each kernel
    activation : str
        Activation function to use
    residual : bool
        Whether to use residual connections
    normalize : Literal["pre", "post", "none"]
        Whether to normalize before or after the residual connection
    """

    def __init__(
        self,
        nlp: Optional[Pipeline] = None,
        name: str = "text_cnn",
        *,
        embedding: WordEmbeddingComponent,
        output_size: Optional[int] = None,
        out_channels: Optional[int] = None,
        kernel_sizes: Sequence[int] = (3, 4, 5),
        activation: ActivationFunction = "relu",
        residual: bool = True,
        normalize: Literal["pre", "post", "none"] = "pre",
    ):
        super().__init__(nlp, name)
        self.embedding = embedding
        self.output_size = output_size or embedding.output_size
        self.module = TextCnn(
            input_size=self.embedding.output_size,
            output_size=self.output_size,
            out_channels=out_channels,
            kernel_sizes=kernel_sizes,
            activation=activation,
            residual=residual,
            normalize=normalize,
        )

    @property
    def span_getter(self):
        return self.embedding.span_getter

    def forward(self, batch: BatchInput) -> WordEmbeddingBatchOutput:
        """
        Encode embeddings with a 1d convolutional network

        Parameters
        ----------
        batch: WordEmbeddingBatchOutput
            - embeddings: embeddings of shape (batch_size, seq_len, input_size)
            - mask: mask of shape (batch_size, seq_len)

        Returns
        -------
        WordEmbeddingBatchOutput
            - embeddings: encoded embeddings of shape (batch_size, seq_len, input_size)
            - mask: (same) mask of shape (batch_size, seq_len)
        """
        embedding_results = self.embedding.module_forward(batch["embedding"])
        if embedding_results["embeddings"].size(0) == 0:
            return embedding_results

        convoluted = self.module(
            embedding_results["embeddings"],
            embedding_results["mask"],
        )
        return {
            "embeddings": convoluted,
            "mask": embedding_results["mask"],
        }



from edsnlp.core import registry

from .diabetes import DiabetesMatcher
from .patterns import default_patterns

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    label="diabetes",
    span_setter={"ents": True, "diabetes": True},
)

create_component = registry.factory.register(
    "eds.diabetes",
    assigns=["doc.ents", "doc.spans"],
)(DiabetesMatcher)



from edsnlp import registry

from .patterns import default_patterns
from .peripheral_vascular_disease import PeripheralVascularDiseaseMatcher

DEFAULT_CONFIG = dict(
    patterns=default_patterns,
    label="peripheral_vascular_disease",
    span_setter={"ents": True, "peripheral_vascular_disease": True},
)

create_component = registry.factory.register(
    "eds.peripheral_vascular_disease",
    assigns=["doc.ents", "doc.spans"],
)(PeripheralVascularDiseaseMatcher)
